import pytest

from steamship import DocTag, Steamship
from steamship.agents.llms import OpenAI
from steamship.agents.schema import AgentContext
from steamship.agents.tools.question_answering import VectorSearchQATool
from steamship.agents.utils import with_llm
from steamship.data import TagKind
from steamship.invocable.mixins.indexer_mixin import IndexerMixin


@pytest.mark.usefixtures("client")
def test_vector_search_qa_tool(client: Steamship):
    """Tests that we can inspect the package and mixin routes"""
    indexer = IndexerMixin(client)
    assert indexer.index_text("Mario was a very fun game.", metadata={"nintendo": True})

    tool = VectorSearchQATool()
    context = with_llm(
        OpenAI(client=client),
        AgentContext.get_or_create(client=client, context_keys={"id": "test"}),
    )
    res = tool.answer_question("What was Mario?", context)
    assert len(res) == 1
    assert "game" in res[0].text

    assert res[0].tags is not None
    sources_tag = None
    for tag in res[0].tags:
        if tag.kind == TagKind.DOCUMENT and tag.name == DocTag.SOURCE:
            sources_tag = tag

    assert sources_tag
    assert sources_tag.value

    assert sources_tag.value.get("sources") is not None
    sources = sources_tag.value.get("sources")
    assert len(sources) == 1

    # This is the metadata we passed in at the top
    assert sources[0].get("nintendo") is True



import logging
import uuid
from typing import List, Optional

from steamship import Block, SteamshipError, Task
from steamship.agents.llms.openai import OpenAI
from steamship.agents.logging import AgentLogging, StreamingOpts
from steamship.agents.schema import Action, Agent, FinishAction
from steamship.agents.schema.context import AgentContext, Metadata
from steamship.agents.utils import with_llm
from steamship.invocable import PackageService, post


class AgentService(PackageService):
    """AgentService is a Steamship Package that can use an Agent, Tools, and a provided AgentContext to
    respond to user input."""

    agent: Optional[Agent]
    """The default agent for this agent service."""

    use_llm_cache: bool
    """Whether or not to cache LLM calls (for tool selection/direct responses) by default."""

    use_action_cache: bool
    """Whether or not to cache agent Actions (for tool runs) by default."""

    max_actions_per_run: int
    """The maximum number of actions to permit while the agent is reasoning.

    This is intended primarily to act as a backstop to prevent a condition in which the Agent decides to loop endlessly
    on tool runs that consume resources with a cost-basis (e.g. prompt completions, embedding operations, vector lookups)
    """

    def __init__(
        self,
        use_llm_cache: Optional[bool] = False,
        use_action_cache: Optional[bool] = False,
        max_actions_per_run: Optional[int] = 5,
        agent: Optional[Agent] = None,
        **kwargs,
    ):
        self.use_llm_cache = use_llm_cache
        self.use_action_cache = use_action_cache
        self.max_actions_per_run = max_actions_per_run
        self.agent = agent
        super().__init__(**kwargs)

    ###############################################
    # Tool selection / execution
    ###############################################

    def next_action(self, agent: Agent, input_blocks: List[Block], context: AgentContext) -> Action:
        action: Action = None
        if context.llm_cache:
            action = context.llm_cache.lookup(key=input_blocks)
        if action:
            logging.info(
                f"Using cached tool selection: calling {action.tool}.",
                extra={
                    AgentLogging.IS_MESSAGE: True,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.THOUGHT,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )
        else:
            logging.info(
                "Selecting next action...",
                extra={
                    AgentLogging.IS_MESSAGE: True,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.THOUGHT,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )
            action = agent.next_action(context=context)
            if context.llm_cache:
                context.llm_cache.update(key=input_blocks, value=action)

        logging.info(
            f"Selected next action: {action.tool}",
            extra={
                AgentLogging.IS_MESSAGE: True,
                AgentLogging.MESSAGE_TYPE: AgentLogging.THOUGHT,
                AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
            },
        )

        return action

    def run_action(self, agent: Agent, action: Action, context: AgentContext):
        if isinstance(action, FinishAction):
            return

        if not agent:
            raise SteamshipError(
                "Missing agent. Not able to run action on behalf of missing agent."
            )

        if context.action_cache:
            # if cache and action is cached, use it. otherwise proceed normally.
            if output_blocks := context.action_cache.lookup(key=action):
                outputs = ",".join([f"{b.as_llm_input()}" for b in output_blocks])
                logging.info(
                    f"Tool {action.tool}: ({outputs}) [cached]",
                    extra={
                        AgentLogging.TOOL_NAME: action.tool,
                        AgentLogging.IS_MESSAGE: True,
                        AgentLogging.MESSAGE_TYPE: AgentLogging.OBSERVATION,
                        AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                    },
                )
                action.output = output_blocks
                context.completed_steps.append(action)
                return

        tool = next((tool for tool in agent.tools if tool.name == action.tool), None)
        if not tool:
            raise SteamshipError(
                f"Could not find tool '{action.tool}' for action. Not able to run."
            )

        # TODO: Arrive at a solid design for the details of this structured log object
        inputs = ",".join([f"{b.as_llm_input()}" for b in action.input])
        logging.info(
            f"Running Tool {action.tool} ({inputs})",
            extra={
                AgentLogging.TOOL_NAME: action.tool,
                AgentLogging.IS_MESSAGE: True,
                AgentLogging.MESSAGE_TYPE: AgentLogging.ACTION,
                AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
            },
        )
        blocks_or_task = tool.run(tool_input=action.input, context=context)
        if isinstance(blocks_or_task, Task):
            raise SteamshipError(
                "Tools return Tasks are not yet supported (but will be soon). "
                "Please use synchronous Tasks (Tools that return List[Block] for now."
            )
        else:
            outputs = ",".join([f"{b.as_llm_input()}" for b in blocks_or_task])
            logging.info(
                f"Tool {action.tool}: ({outputs})",
                extra={
                    AgentLogging.TOOL_NAME: action.tool,
                    AgentLogging.IS_MESSAGE: True,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.OBSERVATION,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )
            action.output = blocks_or_task
            action.is_final = (
                tool.is_final
            )  # Permit the tool to decide if this action should halt the reasoning loop.
            context.completed_steps.append(action)
            if context.action_cache and tool.cacheable:
                context.action_cache.update(key=action, value=action.output)

    def run_agent(self, agent: Agent, context: AgentContext):
        # First, some bookkeeping.

        # Clear any prior agent steps from set of completed steps.
        # This will allow the agent to select tools/dispatch actions based on a new context
        context.completed_steps = []

        # Set the pointer for the current action.
        # The agent will continue to take actions until it is ready to respond.
        action = self.next_action(
            agent=agent, input_blocks=[context.chat_history.last_user_message], context=context
        )

        # Set the counter for the number of actions run.
        # This enables the agent to enforce a budget on actions to guard against running forever.
        number_of_actions_run = 0

        while not action.is_final:
            # If we've exceeded our Action Budget, throw an error.
            if number_of_actions_run >= self.max_actions_per_run:
                raise SteamshipError(
                    message=(
                        f"Agent reached its Action budget of {self.max_actions_per_run} without arriving at a response. If you are the developer, checking the logs may reveal it was selecting unhelpful tools or receiving unhelpful responses from them."
                    )
                )

            # Run the next action and incremenet our counter
            self.run_action(agent=agent, action=action, context=context)
            number_of_actions_run += 1

            # Sometimes, running an action will result in it being dynamically set as a final action as a result of
            # the tool that performed the action's operation. E.g. a Tool wishes to have its output considered the
            # final and complete response.
            #
            # This is a distinct scenario from the next_action method returning a FinalAction, which is why this is
            # a break statement here rather than hidden within the while loop condition above: it's essentially an
            # "early exit" signal.
            if action.is_final:
                break

            # If we're still here, then the Agent has performed work and must decide what to do next. The next
            # action might be another tool, or it might be the Agent deciding to convert this Action's output into
            # A FinalAction object, which would then cause the while loop to terminate upon next iteration.
            action = self.next_action(agent=agent, input_blocks=action.output, context=context)

            # TODO: Arrive at a solid design for the details of this structured log object
            logging.info(
                f"Next Tool: {action.tool}",
                extra={
                    AgentLogging.TOOL_NAME: action.tool,
                    AgentLogging.IS_MESSAGE: False,
                    AgentLogging.MESSAGE_TYPE: AgentLogging.ACTION,
                    AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                },
            )

        context.completed_steps.append(action)
        output_text_length = 0
        if action.output is not None:
            output_text_length = sum([len(block.text or "") for block in action.output])
        logging.info(
            f"Completed agent run. Result: {len(action.output or [])} blocks. {output_text_length} total text length. Emitting on {len(context.emit_funcs)} functions."
        )
        for func in context.emit_funcs:
            logging.info(f"Emitting via function: {func.__name__}")
            func(action.output, context.metadata)

    def set_default_agent(self, agent: Agent):
        self.agent = agent

    def get_default_agent(self, throw_if_missing: bool = True) -> Optional[Agent]:
        """Return the default agent of this agent service.

        This is a helper wrapper to safeguard naming conventions that have formed.
        """
        if hasattr(self, "agent"):
            return self.agent
        elif hasattr(self, "_agent"):
            return self._agent
        else:
            if throw_if_missing:
                raise SteamshipError(
                    message="No Agent object found in the Agent Service. "
                    "Please name it either self.agent or self._agent."
                )
            else:
                return None

    def build_default_context(self, context_id: Optional[str] = None, **kwargs) -> AgentContext:
        """Build's the agent's default context.

        The provides a single place to implement (or override) the default context that will be used by endpoints
        that transports define. This allows an Agent developer to use, eg, the TelegramTransport but with a custom
        type of memory or caching.

        The returned context does not have any emit functions yet registered to it.
        """

        # AgentContexts serve to allow the AgentService to run agents
        # with appropriate information about the desired tasking.
        if context_id is None:
            context_id = uuid.uuid4()
            logging.warning(
                f"No context_id was provided; generated {context_id}. This likely means no conversational history will be present."
            )

        use_llm_cache = self.use_llm_cache
        if runtime_use_llm_cache := kwargs.get("use_llm_cache"):
            use_llm_cache = runtime_use_llm_cache

        use_action_cache = self.use_action_cache
        if runtime_use_action_cache := kwargs.get("use_action_cache"):
            use_action_cache = runtime_use_action_cache

        include_agent_messages = kwargs.get("include_agent_messages", True)
        include_llm_messages = kwargs.get("include_llm_messages", True)
        include_tool_messages = kwargs.get("include_tool_messages", True)

        context = AgentContext.get_or_create(
            client=self.client,
            context_keys={"id": f"{context_id}"},
            use_llm_cache=use_llm_cache,
            use_action_cache=use_action_cache,
            streaming_opts=StreamingOpts(
                include_agent_messages=include_agent_messages,
                include_llm_messages=include_llm_messages,
                include_tool_messages=include_tool_messages,
            ),
        )

        # Add a default LLM to the context, using the Agent's if it exists.
        llm = None
        if agent := self.get_default_agent():
            if hasattr(agent, "llm"):
                llm = agent.llm
        if llm is None:
            llm = OpenAI(client=self.client)

        context = with_llm(context=context, llm=llm)
        return context

    @post("prompt")
    def prompt(
        self, prompt: Optional[str] = None, context_id: Optional[str] = None, **kwargs
    ) -> List[Block]:
        """Run an agent with the provided text as the input."""
        with self.build_default_context(context_id, **kwargs) as context:
            prompt = prompt or kwargs.get("question")
            context.chat_history.append_user_message(prompt)

            # AgentServices provide an emit function hook to access the output of running
            # agents and tools. The emit functions fire at after the supplied agent emits
            # a "FinishAction".
            #
            # Here, we show one way of accessing the output in a synchronous fashion. An
            # alternative way would be to access the final Action in the `context.completed_steps`
            # after the call to `run_agent()`.
            output_blocks = []

            def sync_emit(blocks: List[Block], meta: Metadata):
                nonlocal output_blocks
                output_blocks.extend(blocks)

            context.emit_funcs.append(sync_emit)

            # Get the agent
            agent: Optional[Agent] = self.get_default_agent()
            self.run_agent(agent, context)

            # Now append the output blocks to the chat history
            # TODO: It seems like we've been going from block -> not block -> block here. Opportunity to optimize.
            for output_block in output_blocks:
                # Need to make the output blocks public here so that they can be copied to the chat history.
                # They generally need to be public anyway for the REPL to be able to show a clickable link.
                output_block.set_public_data(True)
                context.chat_history.append_assistant_message(
                    text=output_block.text,
                    tags=output_block.tags,
                    url=output_block.raw_data_url or output_block.url or output_block.content_url,
                    mime_type=output_block.mime_type,
                )

            # Return the response as a set of multi-modal blocks.
            return output_blocks



import pytest
import requests

from steamship import Block, Steamship
from steamship.agents.llms import OpenAI
from steamship.agents.schema import AgentContext
from steamship.agents.tools.image_generation import DalleTool
from steamship.agents.utils import with_llm


@pytest.mark.usefixtures("client")
def test_dalle_tool(client: Steamship):
    """Tests that we can inspect the package and mixin routes"""
    tool = DalleTool()
    context = with_llm(
        OpenAI(client=client),
        AgentContext.get_or_create(client=client, context_keys={"id": "test"}),
    )
    res = tool.run([Block(text="Mario jumping over a fire flower")], context)
    assert len(res) == 1
    assert res[0].public_data is True

    # Test that we can get the public data
    block_url = f"{client.config.api_base}block/{res[0].id}/raw"
    data = requests.get(block_url, json={"id": res[0].id})
    assert data.ok is True
    assert data.content is not None


@pytest.mark.usefixtures("client")
def test_dalle_tool_private_block(client: Steamship):
    tool = DalleTool(make_output_public=False)
    context = with_llm(
        OpenAI(client=client),
        AgentContext.get_or_create(client=client, context_keys={"id": "test"}),
    )
    res = tool.run([Block(text="Mario jumping over a fire flower")], context)
    assert len(res) == 1
    assert res[0].public_data is False



from __future__ import annotations

import logging
import uuid
from logging import StreamHandler
from typing import Dict, List, Optional, Union, cast

from fluent.handler import FluentRecordFormatter

from steamship import File, MimeTypes, Steamship, SteamshipError, Task
from steamship.agents.logging import LOGGING_FORMAT, AgentLogging, StreamingOpts
from steamship.agents.schema.message_selectors import MessageSelector
from steamship.agents.schema.text_splitters import FixedSizeTextSplitter, TextSplitter
from steamship.base.client import Client
from steamship.data import TagKind
from steamship.data.block import Block
from steamship.data.plugin.index_plugin_instance import EmbeddingIndexPluginInstance, SearchResults
from steamship.data.tags import Tag
from steamship.data.tags.tag_constants import ChatTag, DocTag, RoleTag, TagValueKey


class ChatHistory:
    """A ChatHistory is a wrapper of a File ideal for ongoing interactions between a user and a virtual assistant.
    It also includes vector-backed storage for similarity-based retrieval."""

    file: File
    embedding_index: EmbeddingIndexPluginInstance
    text_splitter: TextSplitter

    def __init__(
        self,
        file: File,
        embedding_index: Optional[EmbeddingIndexPluginInstance],
        text_splitter: TextSplitter = None,
    ):
        """This init method is intended only for private use within the class. See `Chat.create()`"""
        self.file = file
        self.embedding_index = embedding_index
        if text_splitter is not None:
            self.text_splitter = text_splitter
        else:
            self.text_splitter = FixedSizeTextSplitter(chunk_size=300)

    @staticmethod
    def _get_existing_file(client: Client, context_keys: Dict[str, str]) -> Optional[File]:
        """Find an existing File object whose memory Tag matches the passed memory keys"""
        file_query = (
            f'kind "{TagKind.CHAT}" and name "{ChatTag.CONTEXT_KEYS}"'
            + (" and " if len(context_keys) > 0 else "")
            + " and ".join([f'value("{key}") = "{value}"' for key, value in context_keys.items()])
        )
        file = File.query(client, file_query)
        if len(file.files) == 1:
            return file.files[0]
        elif len(file.files) == 0:
            return None
        else:
            raise SteamshipError(
                "Multiple ChatHistory objects have been created in this workspace with these memory keys."
            )

    @staticmethod
    def _get_index_handle_from_file(file: File) -> str:
        for tag in file.tags:
            if tag.kind == TagKind.CHAT and tag.name == ChatTag.INDEX_HANDLE:
                return tag.value[TagValueKey.STRING_VALUE]
        raise SteamshipError(f"Could not find index handle on file with id {file.id}")

    @staticmethod
    def _get_embedding_index(client: Steamship, index_handle: str) -> EmbeddingIndexPluginInstance:
        return cast(
            EmbeddingIndexPluginInstance,
            client.use_plugin(
                plugin_handle="embedding-index",
                instance_handle=index_handle,
                config={
                    "embedder": {
                        "plugin_handle": "openai-embedder",
                        "plugin_instance-handle": "text-embedding-ada-002",
                        "fetch_if_exists": True,
                        "config": {"model": "text-embedding-ada-002", "dimensionality": 1536},
                    }
                },
                fetch_if_exists=True,
            ),
        )

    @staticmethod
    def get_or_create(
        client: Steamship,
        context_keys: Dict[str, str],
        tags: List[Tag] = None,
        searchable: bool = True,
    ) -> ChatHistory:

        file = ChatHistory._get_existing_file(client, context_keys)

        if file is None:
            tags = tags or []
            index_handle = str(uuid.uuid4())
            tags.append(Tag(kind=TagKind.DOCUMENT, name=DocTag.CHAT))  # This is a Chat-related tag
            tags.append(Tag(kind=TagKind.CHAT, name=ChatTag.HISTORY))  # This is a ChatHistory file
            tags.append(Tag(kind=TagKind.CHAT, name=ChatTag.CONTEXT_KEYS, value=context_keys))
            tags.append(
                Tag(
                    kind=TagKind.CHAT,
                    name=ChatTag.INDEX_HANDLE,
                    value={TagValueKey.STRING_VALUE: index_handle},
                )
            )

            blocks = []
            file = File.create(
                client=client,
                blocks=blocks,
                tags=tags,
            )
        else:
            index_handle = ChatHistory._get_index_handle_from_file(file)

        if searchable:
            embedding_index = ChatHistory._get_embedding_index(client, index_handle)
        else:
            embedding_index = None

        return ChatHistory(file, embedding_index)

    def append_message_with_role(
        self,
        text: str = None,
        role: RoleTag = RoleTag.USER,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the end-user."""
        tags = tags or []
        tags.append(
            Tag(kind=TagKind.CHAT, name=ChatTag.ROLE, value={TagValueKey.STRING_VALUE: role})
        )
        tags.append(Tag(kind=TagKind.CHAT, name=ChatTag.MESSAGE))
        block = self.file.append_block(
            text=text, tags=tags, content=content, url=url, mime_type=mime_type
        )
        # don't index status messages
        if self.embedding_index is not None and role is not RoleTag.AGENT:
            chunk_tags = self.text_splitter.chunk_text_to_tags(
                block, kind=TagKind.CHAT, name=ChatTag.CHUNK
            )
            block.tags.extend(chunk_tags)
            self.embedding_index.insert(chunk_tags)
        return block

    def append_user_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the end-user."""
        return self.append_message_with_role(text, RoleTag.USER, tags, content, url, mime_type)

    def append_system_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the system, i.e., instructions to the assistant."""
        return self.append_message_with_role(text, RoleTag.SYSTEM, tags, content, url, mime_type)

    def append_assistant_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with content provided by the agent, i.e., results from the assistant."""
        return self.append_message_with_role(text, RoleTag.ASSISTANT, tags, content, url, mime_type)

    @property
    def last_user_message(self) -> Optional[Block]:
        for block in self.file.blocks[::-1]:
            if block.chat_role == RoleTag.USER:
                return block
        return None

    @property
    def last_system_message(self) -> Optional[Block]:
        for block in self.file.blocks[::-1]:
            if block.chat_role == RoleTag.SYSTEM:
                return block
        return None

    @property
    def last_agent_message(self) -> Optional[Block]:
        for block in self.file.blocks[::-1]:
            if block.chat_role == RoleTag.ASSISTANT:
                return block
        return None

    @property
    def initial_system_prompt(self) -> Optional[Block]:
        if len(self.file.blocks) > 0 and self.file.blocks[0].chat_role == RoleTag.SYSTEM:
            return self.file.blocks[0]
        else:
            return None

    def refresh(self):
        self.file.refresh()

    @property
    def tags(self) -> List[Tag]:
        return self.file.tags

    @property
    def messages(self) -> List[Block]:
        return self.file.blocks

    @property
    def client(self) -> Client:
        return self.file.client

    def select_messages(self, selector: MessageSelector) -> List[Block]:
        return selector.get_messages(self.messages)

    def search(self, text: str, k=None) -> Task[SearchResults]:
        if self.embedding_index is None:
            raise SteamshipError("This ChatHistory has no embedding index and is not searchable.")
        return self.embedding_index.search(text, k)

    def is_searchable(self) -> bool:
        return self.embedding_index is not None

    def delete_messages(self, selector: MessageSelector):
        """Delete a set of selected messages from the ChatHistory.

        If `selector == None`, no messages will be deleted.

        NOTES:
        - upon deletion, refresh() is called to ensure up-to-date history refs.
        - causes a full re-index of chat history if the history is searchable.
        """
        if selector:
            selected_messages = selector.get_messages(self.messages)
            for msg in selected_messages:
                msg.delete()

            self.refresh()
            if self.is_searchable():
                self.embedding_index.reset()
                for msg in self.messages:
                    for tag in msg.tags:
                        if tag.kind == TagKind.CHAT and tag.name == ChatTag.CHUNK:
                            # TODO(dougreid): figure out why tag.text gets lost.
                            if not tag.text:
                                tag.text = msg.text[tag.start_idx : tag.end_idx]
                            self.embedding_index.insert(tag)

        self.refresh()

    def clear(self):
        """Deletes ALL messages from the ChatHistory (including system).

        NOTE: upon deletion, refresh() is called to ensure up-to-date history refs.
        """
        for block in self.file.blocks:
            block.delete()

        if self.is_searchable():
            self.embedding_index.reset()

        self.refresh()

    def append_agent_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with status update messages from the Agent."""
        return self.append_message_with_role(text, RoleTag.AGENT, tags, content, url, mime_type)

    def append_tool_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with status update messages from the Agent."""
        return self.append_message_with_role(text, RoleTag.TOOL, tags, content, url, mime_type)

    def append_llm_message(
        self,
        text: str = None,
        tags: List[Tag] = None,
        content: Union[str, bytes] = None,
        url: Optional[str] = None,
        mime_type: Optional[MimeTypes] = None,
    ) -> Block:
        """Append a new block to this with status update messages from the Agent."""
        return self.append_message_with_role(text, RoleTag.LLM, tags, content, url, mime_type)


class ChatHistoryLoggingHandler(StreamHandler):
    """Logs messages emitted by Agents and Tools into a ChatHistory file.

    This is a basic mechanism for streaming status messages alongside generated content.
    """

    chat_history: ChatHistory
    log_level: any
    streaming_opts: StreamingOpts

    def __init__(
        self,
        chat_history: ChatHistory,
        log_level: any = logging.INFO,
        streaming_opts: Optional[StreamingOpts] = None,
    ):
        StreamHandler.__init__(self)
        formatter = FluentRecordFormatter(LOGGING_FORMAT, fill_missing_fmt_key=True)
        self.setFormatter(formatter)
        self.chat_history = chat_history
        self.log_level = log_level
        if streaming_opts is not None:
            self.streaming_opts = streaming_opts
        else:
            self.streaming_opts = StreamingOpts()

    def emit(self, record):
        if record.levelno < self.log_level:
            # don't bother doing anything if level is below logging level
            return

        message_dict = cast(dict, self.format(record))

        is_agent_message = message_dict.get(AgentLogging.MESSAGE_AUTHOR, None) == AgentLogging.AGENT
        if self.streaming_opts.include_agent_messages and is_agent_message:
            return self._append_message(message_dict, AgentLogging.AGENT)

        is_tool_message = message_dict.get(AgentLogging.MESSAGE_AUTHOR, None) == AgentLogging.TOOL
        if self.streaming_opts.include_tool_messages and is_tool_message:
            return self._append_message(message_dict, AgentLogging.TOOL)

        is_llm_message = message_dict.get(AgentLogging.MESSAGE_AUTHOR, None) == AgentLogging.LLM
        if self.streaming_opts.include_llm_messages and is_llm_message:
            return self._append_message(message_dict, AgentLogging.LLM)

    def _append_message(self, message_dict: dict, author_kind: str):
        message = message_dict.get("message", None)
        message_type = message_dict.get(AgentLogging.MESSAGE_TYPE, AgentLogging.MESSAGE)

        if author_kind == AgentLogging.AGENT:
            return self.chat_history.append_agent_message(
                text=message,
                tags=[
                    Tag(
                        kind=TagKind.AGENT_STATUS_MESSAGE,
                        name=message_type,
                        value={TagValueKey.STRING_VALUE: message},
                    ),
                ],
                mime_type=MimeTypes.TXT,
            )
        elif author_kind == AgentLogging.TOOL:
            tool_name = message_dict.get(AgentLogging.TOOL_NAME, AgentLogging.TOOL)
            return self.chat_history.append_tool_message(
                text=message,
                tags=[
                    Tag(
                        kind=TagKind.TOOL_STATUS_MESSAGE,
                        name=message_type,
                        value={TagValueKey.STRING_VALUE: message, "tool": tool_name},
                    )
                ],
                mime_type=MimeTypes.TXT,
            )
        elif author_kind == AgentLogging.LLM:
            llm_name = message_dict.get(AgentLogging.LLM_NAME, AgentLogging.LLM)
            return self.chat_history.append_llm_message(
                text=message,
                tags=[
                    Tag(
                        kind=TagKind.LLM_STATUS_MESSAGE,
                        name=message_type,
                        value={TagValueKey.STRING_VALUE: message, "llm": llm_name},
                    )
                ],
                mime_type=MimeTypes.TXT,
            )



import json
import re
import string
from typing import Dict, List, Optional

from steamship import Block, MimeTypes, Steamship
from steamship.agents.schema import Action, AgentContext, FinishAction, OutputParser, Tool
from steamship.data.tags.tag_constants import RoleTag
from steamship.utils.utils import is_valid_uuid4


def is_punctuation(text: str):
    for c in text:
        if c not in string.punctuation:
            return False
    return True


class FunctionsBasedOutputParser(OutputParser):

    tools_lookup_dict: Optional[Dict[str, Tool]] = None

    def __init__(self, **kwargs):
        tools_lookup_dict = {tool.name: tool for tool in kwargs.pop("tools", [])}
        super().__init__(tools_lookup_dict=tools_lookup_dict, **kwargs)

    def _extract_action_from_function_call(self, text: str, context: AgentContext) -> Action:
        wrapper = json.loads(text)
        fc = wrapper.get("function_call")
        name = fc.get("name", "")
        if name.startswith("functions."):
            name = name[len("functions.") :]  # occasionally, OpenAI prepends "functions."
        tool = self.tools_lookup_dict.get(name, None)
        if tool is None:
            raise RuntimeError(
                f"Could not find tool from function call: `{name}`. Known tools: {self.tools_lookup_dict.keys()}"
            )

        input_blocks = []
        arguments = fc.get("arguments")
        if arguments:
            try:
                args = json.loads(arguments)
                if text := args.get("text"):
                    input_blocks.append(Block(text=text, mime_type=MimeTypes.TXT))
                elif uuid_arg := args.get("uuid"):
                    input_blocks.append(Block.get(context.client, _id=uuid_arg))
            except json.decoder.JSONDecodeError:
                if isinstance(arguments, str):
                    if is_valid_uuid4(arguments):
                        input_blocks.append(Block.get(context.client, _id=uuid_arg))
                    else:
                        input_blocks.append(Block(text=arguments, mime_type=MimeTypes.TXT))

        return Action(tool=tool.name, input=input_blocks, context=context)

    @staticmethod
    def _blocks_from_text(client: Steamship, text: str) -> List[Block]:
        last_response = text.split("AI:")[-1].strip()

        block_id_regex = r"(?:(?:\[|\()?Block)?\(?([A-F0-9]{8}\-[A-F0-9]{4}\-[A-F0-9]{4}\-[A-F0-9]{4}\-[A-F0-9]{12})\)?(?:(\]|\)))?"
        remaining_text = last_response
        result_blocks: List[Block] = []
        while remaining_text is not None and len(remaining_text.strip()) > 0:

            if is_punctuation(remaining_text.strip()):
                remaining_text = ""
                continue

            match = re.search(block_id_regex, remaining_text)
            if match:
                pre_block_text = FunctionsBasedOutputParser._remove_block_prefix(
                    candidate=remaining_text[0 : match.start()]
                )
                if len(pre_block_text) > 0:
                    result_blocks.append(Block(text=pre_block_text))
                result_blocks.append(Block.get(client, _id=match.group(1)))
                remaining_text = FunctionsBasedOutputParser._remove_block_suffix(
                    remaining_text[match.end() :]
                )
            else:
                result_blocks.append(Block(text=remaining_text))
                remaining_text = ""

        return result_blocks

    @staticmethod
    def _remove_block_prefix(candidate: str) -> str:
        removed = candidate
        if removed.endswith("(Block") or removed.endswith("[Block"):
            removed = removed[len("Block") + 1 :]
        elif removed.endswith("Block"):
            removed = removed[len("Block") :]
        return removed

    @staticmethod
    def _remove_block_suffix(candidate: str) -> str:
        removed = candidate
        if removed.startswith(")") or removed.endswith("]"):
            removed = removed[1:]
        return removed

    def parse(self, text: str, context: AgentContext) -> Action:
        if "function_call" in text:
            return self._extract_action_from_function_call(text, context)

        finish_blocks = FunctionsBasedOutputParser._blocks_from_text(context.client, text)
        for finish_block in finish_blocks:
            finish_block.set_chat_role(RoleTag.ASSISTANT)
        return FinishAction(output=finish_blocks, context=context)



from steamship.agents.functional import FunctionsBasedAgent
from steamship.agents.llms.openai import ChatOpenAI
from steamship.agents.schema.message_selectors import MessageWindowMessageSelector
from steamship.agents.service.agent_service import AgentService
from steamship.agents.tools.image_generation import DalleTool
from steamship.agents.tools.search import SearchTool
from steamship.data import TagValueKey
from steamship.invocable import post
from steamship.utils.kv_store import KeyValueStore
from steamship.utils.repl import AgentREPL


class MyCachingAssistant(AgentService):
    """MyCachingAssistant is an example AgentService that exposes a single test endpoint
    for trying out Agent-based invocations. It is configured with two simple Tools
    to provide an overview of the types of tasks it can accomplish (here, search
    and image generation)."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs, use_llm_cache=True, use_action_cache=True)

        # Load the max_actions_per_run from the saved store for use in testing.
        self.kv = KeyValueStore(self.client)
        self.max_actions_per_run = self.get_max_actions_per_run()

        self.set_default_agent(
            FunctionsBasedAgent(
                tools=[
                    SearchTool(),
                    DalleTool(),
                ],
                llm=ChatOpenAI(self.client, temperature=0),
                message_selector=MessageWindowMessageSelector(k=2),
            )
        )

    @post("set_max_actions_per_run")
    def set_max_actions_per_run(self, value: int):
        """Save the max_actions_per_run value so that it will be reloaded upon next request."""
        self.max_actions_per_run = self.kv.set(
            "max_actions_per_run", {TagValueKey.NUMBER_VALUE: value}
        )
        return value

    @post("get_max_actions_per_run")
    def get_max_actions_per_run(self) -> int:
        """Save the max_actions_per_run value so that it will be reloaded upon next request."""
        return (self.kv.get("max_actions_per_run") or {}).get(TagValueKey.NUMBER_VALUE, 5)


if __name__ == "__main__":
    # AgentREPL provides a mechanism for local execution of an AgentService method.
    # This is used for simplified debugging as agents and tools are developed and
    # added.
    AgentREPL(MyCachingAssistant, agent_package_config={}).run()



"""Answers questions with the assistance of a VectorSearch plugin."""
from typing import Any, List, Optional, Union

from steamship import Block, Tag, Task
from steamship.agents.llms import OpenAI
from steamship.agents.schema import AgentContext
from steamship.agents.tools.question_answering.vector_search_tool import VectorSearchTool
from steamship.agents.utils import with_llm
from steamship.utils.repl import ToolREPL


class VectorSearchLearnerTool(VectorSearchTool):
    """Tool to answer questions with the assistance of a vector search plugin."""

    name: str = "VectorSearchLearnerTool"
    human_description: str = "Learns a new fact and puts it in the Vector Database."
    agent_description: str = (
        "Used to remember a fact. Only use this tool if someone asks to remember or learn something. "
        "The input is a fact to learn. "
        "The output is a confirmation that the fact has been learned."
    )

    def learn_sentence(self, sentence: str, context: AgentContext, metadata: Optional[dict] = None):
        """Learns a sigle sentence-sized piece of text.

        GUIDANCE: No more than about a short sentence is a useful unit of embedding search & lookup.
        """
        index = self.get_embedding_index(context.client)
        tag = Tag(text=sentence, metadata=metadata)
        index.insert(tags=[tag])

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        """Learns a fact with the assistance of an Embedding Index plugin.

        Inputs
        ------
        tool_input: List[Block]
            A list of blocks to be rewritten if text-containing.
        context: AgentContext
            The active AgentContext.

        Output
        ------
        output: List[Blocks]
            A lit of blocks containing the answers.
        """

        output = []
        for input_block in tool_input:
            if input_block.is_text():
                self.learn_sentence(input_block.text, context=context)
                output.append(Block(text=f"I'll remember: {input_block.text}"))
        return output


if __name__ == "__main__":
    tool = VectorSearchLearnerTool()
    repl = ToolREPL(tool)

    with repl.temporary_workspace() as client:
        repl.run_with_client(
            client, context=with_llm(context=AgentContext(), llm=OpenAI(client=client))
        )



import time

import pytest
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package
from steamship_tests.utils.fixtures import get_steamship_client

from steamship import File


@pytest.mark.skip(
    reason="article tagging depends on a classifier plugin that needs to be rewritten."
)
def test_article_tagging():
    """NOTE: Any changes to the package under test need to be reflected in the paired documentation pages!"""

    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "article_tagging.py"
    config_template = {
        "labels": {"type": "string"},
    }
    instance_config = {
        "labels": "python,steamship",
    }

    with deploy_package(
        client,
        demo_package_path,
        version_config_template=config_template,
        instance_config=instance_config,
        wait_for_init=True,
    ) as (_, _, instance):
        steamship_file_handle = instance.invoke(
            "add_document", content="steamship, steamship, steamship.", url="https://steamship.com/"
        )
        instance.invoke(
            "add_document",
            content="python is to python as python is to python",
            url="https://www.python.org/",
        )

        # because article tagging doesn't return tasks we can wait on, we retry for ~1m until we find a response.
        tries = 10
        delay = 6  # 6 seconds
        while tries:
            response = instance.invoke("documents_by_tag", tag="steamship")
            if not response:
                tries -= 1
                if not tries:
                    pytest.fail("documents were not tagged within time limit")
                time.sleep(delay)
            else:
                assert len(response) == 1
                assert response[0] == "https://steamship.com/"
                steamship_file = File.get(client, handle=steamship_file_handle)
                assert len(steamship_file.tags[0]) == 2
                return



"""Answers questions with the assistance of a VectorSearch plugin."""
from abc import ABC
from typing import Optional, cast

from steamship import Steamship
from steamship.agents.schema import Tool
from steamship.data.plugin.index_plugin_instance import EmbeddingIndexPluginInstance


class VectorSearchTool(Tool, ABC):
    """Abstract Base Class that provides helper data for a tool that uses Vector Search."""

    embedding_index_handle: Optional[str] = "embedding-index"
    embedding_index_version: Optional[str] = None
    embedding_index_config: Optional[dict] = {
        "embedder": {
            "plugin_handle": "openai-embedder",
            "plugin_instance_handle": "text-embedding-ada-002",
            "fetch_if_exists": True,
            "config": {"model": "text-embedding-ada-002", "dimensionality": 1536},
        }
    }
    embedding_index_instance_handle: str = "default-embedding-index"

    def get_embedding_index(self, client: Steamship) -> EmbeddingIndexPluginInstance:
        index = client.use_plugin(
            plugin_handle=self.embedding_index_handle,
            instance_handle=self.embedding_index_instance_handle,
            config=self.embedding_index_config,
            fetch_if_exists=True,
        )
        return cast(EmbeddingIndexPluginInstance, index)



"""Tool for generating images."""
from steamship import Steamship
from steamship.agents.llms import OpenAI
from steamship.agents.tools import AudioGeneratorTool
from steamship.agents.utils import with_llm
from steamship.utils.repl import ToolREPL


class GenerateSpeechTool(AudioGeneratorTool):
    """Tool to generate audio from text."""

    name: str = "GenerateSpokenAudio"
    human_description: str = "Generates spoken audio from text."
    agent_description: str = (
        "Used to generate spoken audio from text prompts. Only use if the user has asked directly for a "
        "an audio version of output. When using this tool, the input should be a plain text string containing the "
        "content to be spoken."
    )
    generator_plugin_handle: str = "elevenlabs"


if __name__ == "__main__":
    tool = GenerateSpeechTool()
    with Steamship.temporary_workspace() as client:
        ToolREPL(tool).run_with_client(client=client, context=with_llm(llm=OpenAI(client=client)))



import pytest

from steamship import Block, File, MimeTypes, Steamship
from steamship.agents.schema import Action
from steamship.agents.schema.cache import ActionCache, LLMCache


@pytest.mark.usefixtures("client")
def test_llm_cache(client: Steamship):
    context_keys = {"id": "test"}
    cache = LLMCache.get_or_create(client=client, context_keys=context_keys)

    key = [Block(text="prompt")]
    assert not cache.lookup(key)

    cache.update(key=key, value=Action(tool="example_tool", input=[Block(text="input")]))
    action = cache.lookup(key)
    assert action
    assert action.tool == "example_tool"
    assert action.input[0].text == "input"

    cache.update(key=key, value=Action(tool="example_tool", input=[Block(text="other")]))
    action = cache.lookup(key)
    assert action
    assert action.tool == "example_tool"
    assert action.input[0].text == "other"

    cache.delete(key)
    assert not cache.lookup(key)

    file = File.create(client=client, handle="whocares")
    fake_image_block = Block.create(
        client, file_id=file.id, content=bytes([1, 3, 5]), mime_type=MimeTypes.PNG
    )
    text_and_image_key = [Block(text="caption this", mime_type=MimeTypes.TXT), fake_image_block]
    image_action = Action(tool="image_tool", input=text_and_image_key)
    cache.update(key=text_and_image_key, value=image_action)
    action = cache.lookup(key=text_and_image_key)
    assert action
    assert action.tool == "image_tool"
    assert action.input == text_and_image_key

    cache.clear()
    assert not cache.lookup(key=text_and_image_key)


@pytest.mark.usefixtures("client")
def test_action_cache(client: Steamship):
    context_keys = {"id": "test"}
    cache = ActionCache.get_or_create(client=client, context_keys=context_keys)

    file = File.create(client=client, handle="whocares")
    fake_image_block = Block.create(
        client, file_id=file.id, content=bytes([1, 3, 5]), mime_type=MimeTypes.PNG
    )
    text_and_image = [Block(text="caption this", mime_type=MimeTypes.TXT), fake_image_block]
    key = Action(tool="fake_tool", input=text_and_image)
    assert not cache.lookup(key)

    cache.update(key=key, value=text_and_image)
    blocks = cache.lookup(key)
    assert blocks
    assert blocks == text_and_image

    cache.update(key=key, value=[Block(text="other")])
    blocks = cache.lookup(key)
    assert blocks
    assert blocks[0].text == "other"

    cache.delete(key)
    assert not cache.lookup(key)

    cache.update(key=key, value=text_and_image)
    cache.clear()
    assert not cache.lookup(key)



from typing import Any, Dict

from steamship import Block, Steamship
from steamship.agents.mixins.transports.steamship_widget import SteamshipWidgetTransport
from steamship.agents.schema import Action, Agent, AgentContext, FinishAction
from steamship.agents.service.agent_service import AgentService
from steamship.invocable import InvocationContext


class TestAgent(Agent):
    def __init__(self):
        super(TestAgent, self).__init__(tools=[])

    def next_action(self, context: AgentContext) -> Action:
        input_text = context.chat_history.last_user_message.text
        return FinishAction(output=[Block(text=f"Response to: {input_text}")], context=context)


class TestTelegramAgent(AgentService):

    agent: Agent

    def __init__(
        self, client: Steamship, config: Dict[str, Any] = None, context: InvocationContext = None
    ):
        super().__init__(client=client, config=config, context=context)
        self.set_default_agent(TestAgent())
        self.add_mixin(SteamshipWidgetTransport(client=client, agent_service=self))



import pytest
from steamship_tests import PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_plugin

from steamship import Block, Steamship


@pytest.mark.usefixtures("client")
def test_e2e_generator(client: Steamship):
    parser_path = PLUGINS_PATH / "generators" / "test_streaming_generator.py"

    with deploy_plugin(client, parser_path, "generator", streaming=True) as (
        plugin,
        version,
        instance,
    ):
        test_doc = "Yo! Banana boy!"

        # Test unstreamed request
        res = instance.generate(text=test_doc, append_output_to_file=True)
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 1
        assert res.output.blocks[0].text == "!yob ananaB !oY"

        # Test streamed request
        res = instance.generate(text=test_doc, append_output_to_file=True, streaming=True)
        res.wait()
        assert res.output is not None
        assert len(res.output.blocks) == 1
        assert res.output.blocks[0].text == ""  # no text yet!

        streamed_result = str(res.output.blocks[0].raw(), encoding="utf-8")
        assert streamed_result == "!yob ananaB !oY"

        refreshed_block = Block.get(client, _id=res.output.blocks[0].id)
        assert refreshed_block.text == "!yob ananaB !oY"



import pytest
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import MimeTypes, Steamship, Task, TaskState
from steamship.data.plugin.index_plugin_instance import SearchResults


@pytest.mark.usefixtures("client")
def test_indexer_pipeline_mixin(client: Steamship):
    demo_package_path = PACKAGES_PATH / "package_with_mixin_indexer_pipeline.py"

    with deploy_package(client, demo_package_path, wait_for_init=True) as (_, _, instance):
        # Test indexing a pdf file
        pdf_url = "https://steamship.com/test/pdf-test.pdf"

        index_task = instance.invoke("index_url", url=pdf_url)
        index_task = Task.parse_obj(index_task)
        index_task.client = client

        assert index_task.task_id
        assert index_task.state == TaskState.waiting

        index_task.wait()

        result = instance.invoke("search_index", query="education", k=1)
        result = SearchResults.parse_obj(result)
        assert len(result.items) == 1
        winner = result.items[0]
        assert winner.tag.text
        assert winner.tag.value
        assert winner.tag.value.get("block_id")  # It has a block id stamped
        assert winner.tag.value.get("file_id")  # It has a file id stamped
        assert winner.tag.value.get("page") == 0

        # LOAD THE TAO TE QING PDF
        # This will test metadata
        pdf_url2 = "https://www.with.org/tao_te_ching_en.pdf"

        index_task2 = instance.invoke("index_url", url=pdf_url2, metadata={"is_tao": True})
        index_task2 = Task.parse_obj(index_task2)
        index_task2.client = client

        assert index_task2.task_id
        assert index_task2.state == TaskState.waiting

        index_task2.wait()

        result = instance.invoke("search_index", query="Tao", k=1)
        result = SearchResults.parse_obj(result)
        assert len(result.items) == 1
        winner = result.items[0]
        assert winner.tag.text

        assert winner.tag.value
        assert winner.tag.value.get("block_id")  # It has a block id stamped
        assert winner.tag.value.get("file_id")  # It has a file id stamped
        assert winner.tag.value.get("page") is not None
        assert winner.tag.value.get("is_tao") is True
        assert winner.tag.value.get("url") == pdf_url2
        assert winner.tag.value.get("mime_type") == MimeTypes.PDF

        # NOTE NOTE NOTE
        # This portion of the test will be added.. but commented out, to be run only on localhost on an as-needed
        # basis. From experience we've learned that putting YouTube imports in a unit-test is just asking for random
        # failures in the CI/CD pipeline, yet it's still useful to encode what those tests are so that we can debug.

        # youtube_url = "https://www.youtube.com/watch?v=ShPjYHw_K-U"
        #
        # index_task2 = instance.invoke("index_url", url=youtube_url, index_handle="i2")
        # index_task2 = Task.parse_obj(index_task2)
        # index_task2.client = client
        #
        # assert index_task2.task_id
        # assert index_task2.state == TaskState.waiting
        #
        # index_task2.wait()
        #
        # result2 = instance.invoke(
        #     "search_index", query="How do you deploy?", index_handle="i2", k=1
        # )
        # result2 = SearchResults.parse_obj(result2)
        # assert len(result2.items) == 1
        # winner2 = result2.items[0]
        # print(winner2)
        # assert winner2.tag.text



from steamship.agents.functional import FunctionsBasedAgent
from steamship.agents.llms.openai import ChatOpenAI
from steamship.agents.service.agent_service import AgentService
from steamship.agents.tools.question_answering import VectorSearchQATool
from steamship.invocable.mixins.blockifier_mixin import BlockifierMixin
from steamship.invocable.mixins.file_importer_mixin import FileImporterMixin
from steamship.invocable.mixins.indexer_mixin import IndexerMixin
from steamship.invocable.mixins.indexer_pipeline_mixin import IndexerPipelineMixin
from steamship.utils.repl import AgentREPL


class ExampleDocumentQAService(AgentService):
    """DocumentQAService is an example AgentService that exposes:  # noqa: RST201

    - A few authenticated endpoints for learning PDF and YouTube documents:

         /index_url
        { url }

        /index_text
        { text }

    - An unauthenticated endpoint for answering questions about what it has learned

    This agent provides a starter project for special purpose QA agents that can answer questions about documents
    you provide.
    """

    USED_MIXIN_CLASSES = [IndexerPipelineMixin, FileImporterMixin, BlockifierMixin, IndexerMixin]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # This Mixin provides HTTP endpoints that coordinate the learning of documents.
        #
        # It adds the `/learn_url` endpoint which will:
        #    1) Download the provided URL (PDF, YouTube URL, etc)
        #    2) Convert that URL into text
        #    3) Store the text in a vector index
        #
        # That vector index is then available to the question answering tool, below.
        self.add_mixin(IndexerPipelineMixin(self.client, self))

        self.set_default_agent(
            FunctionsBasedAgent(
                tools=[
                    VectorSearchQATool(),  # Tool to answer questions based on a vector store.
                ],
                llm=ChatOpenAI(self.client),
            )
        )


if __name__ == "__main__":
    # AgentREPL provides a mechanism for local execution of an AgentService method.
    # This is used for simplified debugging as agents and tools are developed and
    # added.
    AgentREPL(ExampleDocumentQAService).run()



import logging
import os
import time
from logging import StreamHandler
from typing import Optional, cast

from fluent.handler import FluentRecordFormatter

from steamship.agents.logging import LOGGING_FORMAT, AgentLogging


class DevelopmentLoggingHandler(StreamHandler):
    """A logging handler for developing Steamship Agents, Tools, Packages, and Plugins locally."""

    log_level: any
    log_level_with_message_type: any
    file_log_level: any
    file_handler: Optional[logging.FileHandler]
    log_filename: str

    def __init__(
        self,
        log_level: any = logging.WARN,
        log_level_for_messages: any = logging.INFO,
        file_log_level: any = logging.INFO,
    ):
        StreamHandler.__init__(self)
        formatter = FluentRecordFormatter(LOGGING_FORMAT, fill_missing_fmt_key=True)
        self.setFormatter(formatter)
        self.log_level = log_level
        self.log_level_for_messages = log_level_for_messages
        self.file_log_level = file_log_level

        timestr = time.strftime("shiplog--%Y-%m-%d--%H-%M-%S.log")

        try:
            if not os.path.exists("logs"):
                os.makedirs("logs")
            self.log_filename = os.path.join("logs", timestr)
            self.file_handler = logging.FileHandler(self.log_filename, encoding="utf-8")
        except BaseException:
            print("Unable to create logs/ directory to store debugging logs in current directory.")
            self.file_handler = None

    @staticmethod
    def init_and_take_root(log_level: any = logging.INFO) -> "DevelopmentLoggingHandler":
        logger = logging.getLogger()
        logger.handlers.clear()
        logger.setLevel(log_level)
        dev_logging_handler = DevelopmentLoggingHandler()
        logger.addHandler(dev_logging_handler)
        return dev_logging_handler

    def _emit_regular(self, message_dict: dict):
        level = message_dict.get("level", None)
        message = message_dict.get("message", None)
        print(f"[{level}] {message}")

    def _emit_message(self, message_dict: dict):
        author = message_dict.get(AgentLogging.MESSAGE_AUTHOR, "Unknown")
        message = message_dict.get("message", None)
        message_type = message_dict.get(AgentLogging.MESSAGE_TYPE, AgentLogging.MESSAGE)

        print(f"[{author} {message_type}] {message}")

    def emit(self, record):
        """Emit the record, printing it to console out.

        We rely on TWO logging levels for the mechanics of this LoggingHandler:

        - One for standard logging
        - One for specific system Agent-related events, flagged with metadata

        This is to permit INFO-level logging of key Agent/Tool actions without committing the user to see all
        INFO-level logging globally.

        A future implementation may use a cascade of loggers attached to the AgentContext to do this more cleanly.
        """
        if self.file_handler and record.levelno >= self.file_log_level:
            self.file_handler.emit(record)
            self.file_handler.flush()

        message_dict = cast(dict, self.format(record))

        # It will be returned as a string representation of a bool
        is_message = message_dict.get(AgentLogging.IS_MESSAGE, None) == "True"

        if record.levelno >= self.log_level and not is_message:
            return self._emit_regular(message_dict)
        elif record.levelno >= self.log_level_for_messages and is_message:
            return self._emit_message(message_dict)



import json
import logging
from typing import List, Optional

from steamship import Block, File, PluginInstance, Steamship, Tag
from steamship.agents.logging import AgentLogging
from steamship.agents.schema import LLM, ChatLLM, Tool
from steamship.data import TagKind
from steamship.data.tags.tag_constants import GenerationTag

PLUGIN_HANDLE = "gpt-4"
DEFAULT_MAX_TOKENS = 256


class OpenAI(LLM):
    """LLM that uses Steamship's OpenAI plugin to generate completions.

    NOTE: By default, this LLM uses the `gpt-3.5-turbo` model. Valid model
    choices are `gpt-3.5-turbo` and `gpt-4`.
    """

    generator: PluginInstance
    client: Steamship

    def __init__(
        self, client, model_name: str = "gpt-3.5-turbo", temperature: float = 0.4, *args, **kwargs
    ):
        """Create a new instance.

        Valid model names are:
         - gpt-4
         - gpt-3.5-turbo

        Supported kwargs include:
        - `max_tokens` (controls the size of LLM responses)
        """
        client = client
        max_tokens = DEFAULT_MAX_TOKENS
        if "max_tokens" in kwargs:
            max_tokens = kwargs["max_tokens"]

        generator = client.use_plugin(
            PLUGIN_HANDLE,
            config={"model": model_name, "temperature": temperature, "max_tokens": max_tokens},
        )
        super().__init__(client=client, generator=generator, *args, **kwargs)

    def complete(self, prompt: str, stop: Optional[str] = None, **kwargs) -> List[Block]:
        """Completes the prompt, respecting the supplied stop sequence.

        Supported kwargs include:
        - `max_tokens` (controls the size of LLM responses)
        """
        options = {}
        if stop:
            options["stop"] = stop

        if "max_tokens" in kwargs:
            options["max_tokens"] = kwargs["max_tokens"]

        action_task = self.generator.generate(text=prompt, options=options)
        action_task.wait()
        return action_task.output.blocks


class ChatOpenAI(ChatLLM, OpenAI):
    """ChatLLM that uses Steamship's OpenAI plugin to generate chat completions."""

    def __init__(self, client, model_name: str = "gpt-4-0613", *args, **kwargs):
        """Create a new instance.

        Valid model names are:
         - gpt-4
         - gpt-4-0613

        Supported kwargs include:
        - `max_tokens` (controls the size of LLM responses)
        """
        super().__init__(client=client, model_name=model_name, *args, **kwargs)

    def chat(self, messages: List[Block], tools: Optional[List[Tool]], **kwargs) -> List[Block]:
        """Sends chat messages to the LLM with functions from the supplied tools in a side-channel.

        Supported kwargs include:
        - `max_tokens` (controls the size of LLM responses)
        """

        temp_file = File.create(
            client=self.client,
            blocks=messages,
            tags=[Tag(kind=TagKind.GENERATION, name=GenerationTag.PROMPT_COMPLETION)],
        )

        options = {}
        if len(tools) > 0:
            functions = []
            for tool in tools:
                functions.append(tool.as_openai_function())
            options["functions"] = functions

        if "max_tokens" in kwargs:
            options["max_tokens"] = kwargs["max_tokens"]

        extra = {
            AgentLogging.LLM_NAME: "OpenAI",
            AgentLogging.IS_MESSAGE: True,
            AgentLogging.MESSAGE_TYPE: AgentLogging.PROMPT,
            AgentLogging.MESSAGE_AUTHOR: AgentLogging.LLM,
        }

        if logging.DEBUG >= logging.root.getEffectiveLevel():
            extra["messages"] = json.dumps(
                "\n".join([f"[{msg.chat_role}] {msg.as_llm_input()}" for msg in messages])
            )
            extra["tools"] = ",".join([t.name for t in tools])
        else:
            extra["num_messages"] = len(messages)
            extra["num_tools"] = len(tools)

        logging.info(f"OpenAI ChatComplete ({messages[-1].as_llm_input()})", extra=extra)

        tool_selection_task = self.generator.generate(input_file_id=temp_file.id, options=options)
        tool_selection_task.wait()

        return tool_selection_task.output.blocks



from pydantic.fields import Field
from pydantic.main import BaseModel


class AgentLogging:
    """These keys are for use in the `extra` field of agent logging operations. #noqa: RST203

    For now, they are manually applied at the time of logging. In the future, the AgentContext may provide a logger
    which fills some automatically.

    For example:

    logging.info("I should use tool MakeAPicture", extra={
        AgentLogging.AGENT_NAME: self.name,
        AgentLogging.IS_AGENT_MESSAGE: True,
        AgentLogging.MESSAGE_TYPE: AgentLogging.THOUGHT
    }) # noqa: RST203

    This provides:

    * Structured additions to Fluent/Elastic that help with internal debugging.
    * Helpful output in development mode
    * [Eventual] User-visible logs
    * [Eventual] Visualiations about tool execution and ReAct reasoning

    """

    AGENT_NAME = "agent_name"
    TOOL_NAME = "tool_name"
    LLM_NAME = "llm_name"
    IS_MESSAGE = "is_message"

    MESSAGE_TYPE = "agent_message_type"
    THOUGHT = "thought"
    OBSERVATION = "observation"
    ACTION = "action"
    MESSAGE = "message"
    PROMPT = "prompt"

    MESSAGE_AUTHOR = "message_author"
    USER = "user"
    AGENT = "agent"
    TOOL = "tool"
    SYSTEM = "system"
    LLM = "llm"


LOGGING_FORMAT = {
    "level": "%(levelname)s",
    "host": "%(hostname)s",
    "where": "%(module)s.%(filename)s.%(funcName)s:%(lineno)s",
    "type": "%(levelname)s",
    "stack_trace": "%(exc_text)s",
    "message_type": "%(message_type)s",
    "component_name": "%(component_name)s",
    AgentLogging.IS_MESSAGE: f"%({AgentLogging.IS_MESSAGE})s",  # b doesn't work. Unsure how to make a bool
    AgentLogging.AGENT_NAME: f"%({AgentLogging.AGENT_NAME})s",
    AgentLogging.MESSAGE_AUTHOR: f"%({AgentLogging.MESSAGE_AUTHOR})s",
    AgentLogging.MESSAGE_TYPE: f"%({AgentLogging.MESSAGE_TYPE})s",
    AgentLogging.TOOL_NAME: f"%({AgentLogging.TOOL_NAME})s",
}


class StreamingOpts(BaseModel):
    """Controls what status messages for a given AgentService invocation should be included in the resultant stream."""

    include_agent_messages: bool = Field(default=True)
    """Whether or not to include agent-generated messages in the ChatHistory stream.

    Agent messages are any logging messages with: `AgentLogging.MESSAGE_AUTHOR == AgentLogging.AGENT`. These typically
    are status messages logged in either the `AgentService` or the `Agent` code.
    """

    include_llm_messages: bool = Field(default=True)
    """Whether or not to include LLM-generated messages in the ChatHistory stream.

    LLM messages are any logging messages with: `AgentLogging.MESSAGE_AUTHOR == AgentLogging.LLM`. These typically
    are status messages logged in LLM code (ex: `ChatOpenAI`).
    """

    include_tool_messages: bool = Field(default=True)
    """Whether or not to include Tool-generated messages in the ChatHistory stream.

    LLM messages are any logging messages with: `AgentLogging.MESSAGE_AUTHOR == AgentLogging.TOOL`. These typically
    are status messages logged in Tool code (ex: `GeneratorTool`).
    """

    @property
    def stream_intermediate_events(self):
        return (
            self.include_llm_messages or self.include_agent_messages or self.include_tool_messages
        )



import json
from typing import List

import pytest
from steamship_tests import TEST_ASSETS_PATH
from steamship_tests.utils.deployables import zip_deployable

from steamship import Package, PackageInstance, PackageVersion, Steamship, SteamshipError
from steamship.data.package.package_version import CreatePackageVersionRequest
from steamship.utils.url import Verb


@pytest.mark.usefixtures("client")
def test_install_two_packages_at_once(client: Steamship):

    pillow_version_task = create_version(client, additional_requirements=["pillow"])
    pandas_version_task = create_version(client, additional_requirements=["pandas"])

    pillow_version = pillow_version_task.wait()
    pandas_version = pandas_version_task.wait()

    pillow_instance = PackageInstance.create(
        client, package_id=pillow_version.package_id, package_version_id=pillow_version.id
    )
    pandas_instance = PackageInstance.create(
        client, package_id=pandas_version.package_id, package_version_id=pandas_version.id
    )

    pillow_instance.wait_for_init()
    pandas_instance.wait_for_init()

    # Pillow version should be able to run the method that imports pillow locally
    assert pillow_instance.invoke("try_pillow", verb=Verb.GET) == "PIL"

    # Pillow version should NOT be able to run the method that imports pandas locally
    with pytest.raises(SteamshipError):
        _ = pillow_instance.invoke("try_pandas", verb=Verb.GET)

    # Pandas version should be able to run the method that imports pandas locally
    assert pandas_instance.invoke("try_pandas", verb=Verb.GET) == "pandas"

    # Pandas version should NOT be able to run the method that imports pillow locally
    with pytest.raises(SteamshipError):
        _ = pandas_instance.invoke("try_pillow", verb=Verb.GET)

    pillow_version.delete()
    pandas_version.delete()


def create_version(client: Steamship, additional_requirements: List[str]):
    """Separate implementation of create version so that we can not wait on the task, and run multiple at once."""
    package = Package.create(client)

    zip_bytes = zip_deployable(
        TEST_ASSETS_PATH / "packages" / "requirement_isolation_package.py",
        additional_requirements=additional_requirements,
    )
    hosting_handler = "steamship.invocable.entrypoint.safe_handler"

    req = CreatePackageVersionRequest(
        handle="handle",
        package_id=package.id,
        config_template=json.dumps({}),
        hosting_handler=hosting_handler,
    )

    task = client.post(
        "package/version/create",
        payload=req,
        file=("package.zip", zip_bytes, "multipart/form-data"),
        expect=PackageVersion,
    )
    return task



import json
import logging
import os
import platform
import signal
import subprocess  # noqa: S404
import sys
import time
from os import getenv, path
from typing import Optional

import click

import steamship
from steamship import PackageInstance, Steamship, SteamshipError, Workspace
from steamship.base.configuration import DEFAULT_WEB_BASE, Configuration
from steamship.cli.create_instance import (
    config_str_to_dict,
    create_instance,
    load_manifest,
    set_unset_params,
)
from steamship.cli.deploy import (
    PackageDeployer,
    PluginDeployer,
    bundle_deployable,
    update_config_template,
)
from steamship.cli.local_server.server import SteamshipHTTPServer
from steamship.cli.manifest_init_wizard import manifest_init_wizard
from steamship.cli.requirements_init_wizard import requirements_init_wizard
from steamship.cli.ship_spinner import ship_spinner
from steamship.cli.utils import find_api_py, get_api_module
from steamship.data.manifest import DeployableType, Manifest
from steamship.data.user import User
from steamship.invocable.dev_logging_handler import DevelopmentLoggingHandler
from steamship.invocable.lambda_handler import get_class_from_module
from steamship.utils.repl import HttpREPL


@click.group()
def cli():
    pass


def initialize(suppress_message: bool = False):
    logging.root.setLevel(logging.FATAL)
    if not suppress_message:
        click.echo(f"Steamship Python CLI version {steamship.__version__}")


def initialize_and_get_client_and_prep_project():
    initialize()
    client = None
    try:
        client = Steamship()
    except SteamshipError as e:
        click.secho(e.message, fg="red")
        click.get_current_context().abort()

    user = User.current(client)
    if path.exists("steamship.json"):
        manifest = Manifest.load_manifest()
    else:
        manifest = manifest_init_wizard(client)
        manifest.save()

    if not path.exists("requirements.txt"):
        requirements_init_wizard()

    update_config_template(manifest)

    return client, user, manifest


@click.command()
def login():
    """Log in to Steamship, creating ~/.steamship.json"""
    initialize()
    click.echo("Logging into Steamship.")
    if sys.argv[1] == "login":
        if Configuration.default_config_file_has_api_key():
            overwrite = click.confirm(
                text="You already have an API key in your .steamship.json file.  Do you want to remove it and login?",
                default=False,
            )
            if not overwrite:
                sys.exit(0)
            Configuration.remove_api_key_from_default_config()

        # Carry on with login
        client = Steamship()
        user = User.current(client)
        click.secho(f"🚢🚢🚢 Hooray! You're logged in with user handle: {user.handle} 🚢🚢🚢", fg="green")


@click.command()
def ships():
    """Ship some ships"""
    initialize()
    click.secho("Here are some ships:", fg="cyan")
    with ship_spinner():
        time.sleep(5)
    click.secho()


def _run_ngrok(local_port: int) -> str:
    """Create an NGROK URL directed at `local_port`."""
    try:
        from pyngrok import ngrok
    except BaseException:
        click.secho("⚠️ Public API: Unable to start ngrok. Please either:")
        click.secho("   - Install pyngrok via `pip install pyngrok`")
        click.secho("   - Use the --no-ngrok flag")
        click.secho("   NGROK is only necessary if you wish to debug Telegram or Slack locally.")
        exit(1)
    try:
        http_tunnel = ngrok.connect(local_port, bind_tls=True)
    except BaseException:
        click.secho(f"⚠️ Public API: Unable to bind ngrok to port {local_port}")
        click.secho("   - Try running with a different port via the  --port flag.")
        exit(1)

    ngrok_api_url = http_tunnel.public_url
    return ngrok_api_url


def _run_local_server(
    local_port: int,
    instance_handle: Optional[str] = None,
    config: Optional[str] = None,
    workspace: Optional[str] = None,
    base_url: Optional[
        str
    ] = None,  # If provided, will override the default calculation, eg for ngrok
) -> str:
    """Start the local API Server."""
    logging.info(
        f"Starting local server. port={local_port}, instance_handle={instance_handle}, workspace={workspace}"
    )

    path = find_api_py()
    api_module = get_api_module(path)
    invocable_class = get_class_from_module(api_module)

    # Use the provided base url (e.g. from NGROK) or default to localhost
    _base_url = base_url or "http://localhost"

    logging.info(f"Local server base URL (for PackageService configuration) is {_base_url}")

    if not invocable_class:
        logging.error("Local server startup unable to find Steamship service.")
        click.secho("⚠️ Local API: Unable to find Steamship service. Please:")
        click.secho(
            "   - Check to see that you have an api.py file containing an AgentService or PackageService "
        )
        exit(1)

    manifest = load_manifest()

    if not manifest:
        logging.error("Local server startup unable to find Steamship manifest.")
        click.secho("⚠️ Local API: Unable to find your steamship.json file")
        exit(1)

    invocable_config, is_file = config_str_to_dict(config)
    set_unset_params(config, invocable_config, is_file, manifest)

    server = SteamshipHTTPServer(
        invocable_class,
        base_url=_base_url,
        port=local_port,
        invocable_handle=manifest.handle,
        invocable_version_handle=manifest.version,
        invocable_instance_handle=instance_handle,
        config=invocable_config,
        workspace=workspace,
    )
    try:
        server.start()
    except SteamshipError as e:
        click.secho(f"⚠️ Local API: {e.message}")
        if "validation error" in e.message:
            click.secho(
                "💡 Suggestion: Create a configuration file called config.json containing the required "
            )
            click.secho(
                "               configuration fields in your agent's configuration. Then run again with"
            )
            click.secho("               ship run local -c config.json")
            exit(-1)
    except BaseException as e:
        click.secho(f"⚠️ Local API: {e}")
        exit(-1)

    logging.info("Local server running.")

    def on_exit(signum, frame):
        logging.info("Local server shutting down.")
        click.secho("Shutting down server.")
        server.stop()
        click.secho("Shut down.")
        logging.info("Local server shut down complete.")
        exit(1)

    signal.signal(signal.SIGINT, on_exit)

    local_api_url = f"http://localhost:{server.port}"

    logging.info(f"Local server address: {local_api_url}")

    return local_api_url


def _run_web_interface(base_url: str) -> str:
    web_base = DEFAULT_WEB_BASE
    logging.info(f"Starting web interface. web_base={web_base} and base_url={base_url}")

    try:
        config = Configuration()
        web_base = config.web_base
    except BaseException:
        click.secho("⚠️ Web UI:  Unable to find Steamship Configuration. Please:")
        click.secho(
            "   - Run `ship login` to make sure you have Steamship credentials in your environment."
        )

    # Guarantee one and only one trailing /
    if not web_base.endswith("/"):
        web_base = f"{web_base}/"

    web_url = f"{web_base}debug?endpoint={base_url}/prompt"
    logging.info(f"Web interface url is: {web_url}")

    return web_url


def register_locally_running_package_with_engine(
    client: Steamship,
    ngrok_api_url: str,
    package_handle: str,
    manifest: Manifest,
    config: Optional[str] = None,
) -> PackageInstance:
    """Registers the locally running package with the Steamship Engine."""

    # Register the Instance in the Engine
    invocable_config, is_file = config_str_to_dict(config)
    set_unset_params(config, invocable_config, is_file, manifest)
    package_instance = PackageInstance.create_local_development_instance(
        client,
        local_development_url=ngrok_api_url,
        package_handle=package_handle,
        config=invocable_config,
    )
    return package_instance


def serve_local(  # noqa: C901
    port: int = 8443,
    no_ngrok: Optional[bool] = False,
    no_repl: Optional[bool] = False,
    no_ui: Optional[bool] = False,
    config: Optional[str] = None,
    workspace: Optional[str] = None,
):
    """Serve the invocable on localhost. Useful for debugging locally."""
    dev_logging_handler = DevelopmentLoggingHandler.init_and_take_root()

    click.secho("Running your project...\n")

    client, user, manifest = initialize_and_get_client_and_prep_project()

    if workspace:
        workspace_obj = Workspace.get(client, handle=workspace)
    else:
        workspace_obj = Workspace.get(client)
        workspace = workspace_obj.handle

    # Make sure we're running a package.
    if manifest.type != DeployableType.PACKAGE:
        click.secho(
            f"⚠️ Must run `ship serve local` in a folder with a Steamship Package. Found: {manifest.type}"
        )
        exit(-1)

    # Make sure we have a package name -- this allows us to register the running copy with the engine.
    deployer = PackageDeployer()
    deployable = deployer.create_or_fetch_deployable(client, user, manifest)

    # Report the logs output file.
    click.secho(f"📝 Log file:   {dev_logging_handler.log_filename}")

    # Start the NGROK connection
    ngrok_api_url = None
    public_api_url = None

    # Hard coded instance handle to represent "a local instance that isn't connected to the engine"
    local_instance_handle = "local-dev-instance-not-connected-to-engine"

    if not no_ngrok:
        ngrok_api_url = _run_ngrok(port)

        # It requires a trailing slash
        if ngrok_api_url[-1] != "/":
            ngrok_api_url = ngrok_api_url + "/"

        registered_instance = register_locally_running_package_with_engine(
            client=client,
            ngrok_api_url=ngrok_api_url,
            package_handle=deployable.handle,
            manifest=manifest,
            config=config,
        )

        # Replace the local instance handle with the instance just registered in the engine.
        local_instance_handle = registered_instance.handle

        # Notes:
        #  1. registered_instance.invocation_url is the NGROK URL, not the Steamship Proxy URL.
        #  2. The public_api_url should still be NGROK, not the Proxy. The local server emulates the Proxy and
        #     the Proxy blocks this kind of development traffic.

        public_api_url = ngrok_api_url
        click.secho(f"🌎 Public API: {public_api_url}")

    # Start the local API Server. This has to happen after NGROK because the port & url need to be plummed.
    try:
        local_api_url = _run_local_server(
            local_port=port,
            instance_handle=local_instance_handle,
            config=config,
            workspace=workspace,
            base_url=public_api_url,
        )
    except BaseException as e:
        click.secho("⚠️ Local API:  Unable to start local server.")
        click.secho(e)
        exit(-1)

    if local_api_url[-1] != "/":
        local_api_url = local_api_url + "/"

    if local_api_url:
        click.secho(f"🌎 Local API:  {local_api_url}")
    else:
        click.secho("⚠️ Local API:  Unable to start local server.")
        exit(-1)

    # Start the web UI
    if not no_ui:
        web_url = _run_web_interface(public_api_url or local_api_url)
        if web_url:
            click.secho(f"🌎 Web UI:     {web_url}")

    if no_repl:
        while True:
            # We need to make sure the thread doesn't exit.
            time.sleep(1)
    else:
        click.secho("\n💬 Interactive REPL below. Type to interact.\n")
        prompt_url = f"{local_api_url or public_api_url}prompt"
        repl = HttpREPL(prompt_url=prompt_url, dev_logging_handler=dev_logging_handler)
        repl.run()


@click.command()
@click.option(
    "--port",
    "-p",
    type=int,
    default=8443,
    help="Port to host the server on.",
)
@click.option(
    "--instance_handle",
    "-h",
    type=str,
    default=None,
    help="Handle of the package or plugin instance being hosted.",
)
@click.option(
    "--no-ngrok",
    is_flag=True,
    default=False,
    help="Don't attempt to attach to ngrok.",
)
@click.option(
    "--no-ui",
    is_flag=True,
    default=False,
    help="Don't attempt to attach to a web UI.",
)
@click.option(
    "--no-repl",
    is_flag=True,
    default=False,
    help="Don't start a console REPL.",
)
@click.option(
    "--config",
    "-c",
    type=str,
    required=False,
    help="Instance configuration. May be inline JSON or a path to a file. If not specified, "
    "an empty configuration dictionary will be passed to the instance.",
)
@click.option(
    "--workspace",
    "-w",
    required=False,
    type=str,
    help="Workspace handle. The new instance will be created in this workspace. If not specified, "
    "the default workspace will be used.",
)
@click.argument(
    "environment", required=True, type=click.Choice(["local", "remote"], case_sensitive=False)
)
@click.pass_context
def run(
    ctx,
    environment: str,
    port: int = 8080,
    instance_handle: Optional[str] = None,
    no_ngrok: Optional[bool] = False,
    no_ui: Optional[bool] = False,
    no_repl: Optional[bool] = False,
    config: Optional[str] = None,
    workspace: Optional[str] = None,
):
    """Serve your invocable locally or on prod"""
    if environment == "local":
        serve_local(
            port=port,
            no_ngrok=no_ngrok,
            no_repl=no_repl,
            no_ui=no_ui,
            config=config,
            workspace=workspace,
        )
    else:
        if click.confirm("Do you want to deploy a new version first?", default=False):
            ctx.invoke(deploy)
        ctx.invoke(
            create_instance, workspace=workspace, instance_handle=instance_handle, config=config
        )


@click.command()
def deploy():
    """Deploy the package or plugin in this directory"""
    client, user, manifest = initialize_and_get_client_and_prep_project()

    deployable_type = manifest.type

    deployer = None
    if deployable_type == DeployableType.PACKAGE:
        deployer = PackageDeployer()
    elif deployable_type == DeployableType.PLUGIN:
        deployer = PluginDeployer()
    else:
        click.secho("Deployable must be of type package or plugin.", fg="red")
        click.get_current_context().abort()

    deployable = deployer.create_or_fetch_deployable(client, user, manifest)

    click.echo("Bundling content... ", nl=False)
    bundle_deployable(manifest)
    click.echo("Done. 📦")

    _ = deployer.create_version(client, manifest, deployable.id)

    thing_url = f"{client.config.web_base}{deployable_type.value}s/{manifest.handle}"
    click.echo(
        f"Deployment was successful. View and share your new {deployable_type.value} here:\n\n{thing_url}\n"
    )

    # Common error conditions:
    # - Package/plugin handle already taken. [handled; asks user for new]
    # - Version handle already deployed. [handled; asks user for new]
    # - Bad parameter configuration. [mitigated by deriving template from Config object]
    # - Package content fails health check (ex. bad import) [Error caught while checking config object]


@click.command()
def info():
    """Displays information about the current Steamship user.

    This is useful to help users (in a support or hackathon context) test whether they have configured their
    Steamship environment correctly.
    """
    initialize()
    click.echo("\nSteamship Client Info\n=====================\n")

    if Configuration.default_config_file_has_api_key() or getenv("STEAMSHIP_API_KEY", None):
        # User is logged in!
        client = None
        try:
            client = Steamship()
        except BaseException:
            click.secho("Incorrect API key or network error.\n", fg="red")
            click.secho(
                "Your Steamship API Key is set, but we were unable to use it to fetch your account information.\n"
            )
            click.secho("- If you are on your own computer, run `ship login` to login.")
            click.secho(
                "- If you are in Replit, add the STEAMSHIP_API_KEY secret, then close and re-open this shell.\n"
            )
            return

        try:
            user = User.current(client)
            click.echo(f"User handle: {user.handle}")
            click.echo(f"User ID:     {user.id}")
            click.echo(f"Profile:     {client.config.profile}")
            click.echo("\nReady to ship! 🚢🚢🚢\n")

        except BaseException:
            click.secho("Incorrect API key or network error.\n", fg="red")
            click.secho(
                "Your Steamship API Key is set, but we were unable to use it to fetch your account information.\n"
            )
            click.secho("- If you are on your own computer, run `ship login` to login.")
            click.secho(
                "- If you are in Replit, add the STEAMSHIP_API_KEY secret, then close and re-open this shell.\n"
            )
    else:
        click.secho("You are not logged in.\n")
        click.secho("- If you are on your own computer, run `ship login` to login.")
        click.secho(
            "- If you are in Replit, add the STEAMSHIP_API_KEY secret, then close and re-open this shell.\n"
        )


@click.command()
@click.option(
    "--workspace",
    "-w",
    required=True,
    type=str,
    help="Workspace handle used for scoping logs request. All requests MUST be scoped by workspace.",
)
@click.option(
    "--offset",
    "-o",
    default=0,
    type=int,
    help="Starting index from sorted logs to return a chunk. Used for paging. Defaults to 0.",
)
@click.option(
    "--number",
    "-n",
    default=50,
    type=int,
    help="Number of logs to return in a single batch. Defaults to 50.",
)
@click.option(
    "--package",
    "-p",
    type=str,
    help="Package handle. Used to filter logs returend to a specific package (across all instances).",
)
@click.option(
    "--instance",
    "-i",
    type=str,
    help="Instance handle. Used to filter logs returned to a specific instance of a package.",
)
@click.option(
    "--version",
    "-v",
    type=str,
    help="Version handle. Used to filter logs returned to a specific version of a package.",
)
@click.option(
    "--path",
    "request_path",
    type=str,
    help="Path invoked by a client operation. Used to filter logs returned to a specific invocation path.",
)
@click.option(
    "--with-fields",
    "-f",
    "field_values",
    type=str,
    help="Dictionary of log field values (format: key1=value1,key2=value2,...). Used to filter logs returned.",
)
def logs(
    workspace: str,
    offset: int,
    number: int,
    package: Optional[str] = None,
    instance: Optional[str] = None,
    version: Optional[str] = None,
    request_path: Optional[str] = None,
    field_values: Optional[str] = None,
):
    """Retrieve logs within a workspace."""

    initialize(suppress_message=True)
    client = None
    try:
        client = Steamship(workspace=workspace)
    except SteamshipError as e:
        raise click.UsageError(message=e.message)

    value_dict = {}
    if field_values:
        try:
            for item in field_values.split(","):
                key, value = item.split("=")
                value_dict[key] = value
        except ValueError:
            raise click.UsageError(
                message="Invalid dictionary format for fields. Please provide a dictionary in the "
                "format: key1=value1,key2=value2,..."
            )

    click.echo(
        json.dumps(
            client.logs(offset, number, package, instance, version, request_path, value_dict)
        )
    )


def _exit_if_not_proceed(skip: bool, prompt: str):
    if skip:
        return

    click.secho("⚠️ Deletion is a destructive operation. It is not recoverable. ⚠️", fg="red")
    click.secho(prompt, fg="red")
    confirm = click.confirm("Proceed?", default=False)
    if not confirm:
        exit(0)


@click.command()
@click.option(
    "--workspace",
    "-w",
    required=True,
    type=str,
    help="Workspace handle.",
)
@click.option(
    "--instance",
    "-i",
    type=str,
    help="Instance handle.",
)
@click.option("--yes", is_flag=True)
def delete(
    workspace: str,
    instance: Optional[str] = None,
    yes: Optional[bool] = False,
):
    """Deletes Steamship workspaces and instances, based on provided fields.

    NOTE: If `instance` is not specified, the `workspace` will be deleted.
    """
    initialize(suppress_message=True)
    client = None
    try:
        client = Steamship(workspace=workspace)
    except SteamshipError as e:
        raise click.UsageError(message=e.message)

    if not instance:
        # delete workspace
        _exit_if_not_proceed(
            skip=yes,
            prompt=f"This will delete workspace '{workspace}' and all data contained within.",
        )
        wspace = client.get_workspace()
        click.secho(f"Deleting workspace '{workspace}'... ", nl=False, fg="green")
        wspace.delete()
        click.secho("Done.", fg="green")
        return

    if instance:
        # delete instance
        _exit_if_not_proceed(
            skip=yes, prompt=f"This will delete instance '{instance}' in workspace '{workspace}'"
        )
        try:
            pkg_inst = PackageInstance.get(client=client, handle=instance)
            click.secho(
                f"Deleting instance '{instance}' in workspace '{workspace}'... ",
                nl=False,
                fg="green",
            )
            pkg_inst.delete()
            click.secho("Done.", fg="green")
            return
        except SteamshipError as e:
            click.secho(
                f"⚠️ Failed to delete instance '{instance}' in workspace '{workspace}': {e.message}",
                fg="red",
            )
            exit(1)


@click.command()
def support_info():
    """Displays detailed information needed for getting technical support"""
    initialize()
    click.echo("\nSteamship User Info\n=====================")

    if Configuration.default_config_file_has_api_key() or getenv("STEAMSHIP_API_KEY", None):
        # User is logged in!
        client = None
        try:
            client = Steamship()
            user = User.current(client)
            click.echo(f"User ID:     {user.id}")

        except BaseException:
            click.secho("User not logged in.")
    else:
        click.secho("User not logged in.")

    click.echo("\n\nDeployable Info\n=====================")
    if path.exists("steamship.json"):
        manifest = Manifest.load_manifest()
        click.echo(f"Deployable type: {manifest.type.value}")
        if manifest.type.value == DeployableType.PLUGIN:
            click.echo(f"Plugin type: {manifest.plugin.type}")
        click.echo(f"Deployable handle: {manifest.handle}")
        click.echo(f"Deployable version: {manifest.version}")

    else:
        click.echo("No deployable manifest.")
    click.echo("\n\nDependency Info\n=====================")
    click.echo(f"Running Python CLI version: {steamship.__version__}")
    click.echo("\nEnv packages:")
    click.echo("\n-------------")
    subprocess.run(["pip", "list"])  # noqa: S607, S603
    click.echo("\nRequirements.txt:")
    click.echo("\n-----------------")
    if path.exists("requirements.txt"):
        with open("requirements.txt") as requirements:
            lines = requirements.readlines()
            if len(lines) == 0:
                click.secho("Empty requirements.txt")
            else:
                for line in lines:
                    click.secho(line, nl=False)
                click.secho("")
    else:
        click.echo("FILE NOT PRESENT")

    click.echo("\n\nEnvironment/OS Info\n=====================")
    click.secho(f"OS type: {os.name}")
    click.secho(f"OS Name: {platform.system()}")
    click.secho(f"OS Version: {platform.release()}")
    click.secho(f"Python version: {sys.version}")
    click.secho(f"Shell: {os.environ.get('SHELL')}")
    click.secho(f"In virtual env: {sys.prefix != sys.base_prefix}")


cli.add_command(login)
cli.add_command(deploy)
cli.add_command(info)
cli.add_command(deploy, name="it")
cli.add_command(ships)
cli.add_command(logs)
cli.add_command(run)
cli.add_command(create_instance, name="use")
cli.add_command(delete)
cli.add_command(support_info, name="support-info")



from pathlib import Path

from steamship_tests import PACKAGES_PATH, PLUGINS_PATH
from steamship_tests.utils.deployables import zip_deployable

output_path = "../../nludb/TestAssets/"


def build_asset(py_path: Path, output_file: str):
    zip_bytes = zip_deployable(py_path)
    with open(output_path + output_file, "wb") as f:
        f.write(zip_bytes)


def main():

    assets_to_build = [
        (PLUGINS_PATH / "blockifiers" / "csv_blockifier.py", "csv-blockifier.zip"),
        (PACKAGES_PATH / "demo_package.py", "demo-package.zip"),
        (PACKAGES_PATH / "configurable_hello_world.py", "configurable_hello_world.zip"),
        (PLUGINS_PATH / "blockifiers" / "blockifier.py", "dummy-blockifier.zip"),
        (PACKAGES_PATH / "bad_package.py", "bad-package.zip"),
        (PACKAGES_PATH / "package_with_instance_init.py", "package-with-instance-init.zip"),
        (
            PACKAGES_PATH / "package_with_failing_instance_init.py",
            "package-with-failing-instance-init.zip",
        ),
        (
            PLUGINS_PATH / "generators" / "plugin_with_instance_init.py",
            "plugin-with-instance-init.zip",
        ),
        (
            PLUGINS_PATH / "generators" / "request_id_generator.py",
            "request-id-generator.zip",
        ),
        (
            PACKAGES_PATH / "request_id_plumbing_test_package.py",
            "request-id-plumbing-test-package.zip",
        ),
    ]

    for path, output in assets_to_build:
        build_asset(path, output)


main()



import pytest
from steamship_tests import PACKAGES_PATH, PLUGINS_PATH
from steamship_tests.utils.deployables import deploy_package, deploy_plugin

from steamship import Steamship


@pytest.mark.usefixtures("client")
def test_get_param_decoding(client: Steamship):
    request_id_package_path = PACKAGES_PATH / "request_id_plumbing_test_package.py"
    request_id_plugin_path = PLUGINS_PATH / "generators" / "request_id_generator.py"

    with deploy_package(client, request_id_package_path) as (package, version, instance):
        with deploy_plugin(client, request_id_plugin_path, plugin_type="generator") as (
            plugin,
            _,
            _,
        ):
            result = instance.invoke("requestids", plugin_handle=plugin.handle)
            package_request_id = result.get("packageRequestId")
            plugin_request_id = result.get("pluginRequestId")
            invoke_later_request_id = result.get("invokeLaterRequestId")
            assert package_request_id is not None
            assert package_request_id == plugin_request_id
            assert package_request_id == invoke_later_request_id



import json
import logging
import urllib.parse
from enum import Enum
from typing import List, Optional

import requests
from pydantic import BaseModel, Field

from steamship import Block, Steamship
from steamship.agents.llms import OpenAI
from steamship.agents.mixins.transports.transport import Transport
from steamship.agents.schema import EmitFunc, Metadata
from steamship.agents.service.agent_service import AgentService
from steamship.agents.utils import with_llm
from steamship.invocable import Config, InvocableResponse, get, post
from steamship.utils.kv_store import KeyValueStore

SLACK_API_BASE = "https://slack.com/api/"
SETTINGS_KVSTORE_KEY = "slack-transport"


class SlackContextBehavior(Enum):
    """Defines how history between agent and users is tracked.

    These specifications are specifically in regard to how the agent interacts with Slack as it pertains to Agent
    Context.
    """

    ENTIRE_CHANNEL = "entire-channel"
    """
    Agent context is per channel as a whole, which includes bot mentions sent to the top level channel, and across *any*
    thread in that channel.
    """

    THREADS_ARE_NEW_CONVERSATIONS = "threads-are-new-conversations"
    """
    Agent context is thread-aware. The top level channel is treated as its own context, and threads have their own
    contexts.
    """


class SlackThreadingBehavior(Enum):
    """Defines how responses from the agent will be delivered in response to user mentions."""

    FOLLOW_THREADS = "follow-threads"
    """
    If the bot is mentioned from the top-level channel, the response will be in the channel. If the bot is mentioned
    from within a thread, the response will be to that thread.
    """

    ALWAYS_THREADED = "always-threaded"
    """
    Responses from the bot will always be threaded. If the bot was mentioned at the top level of the channel, a new
    thread will be created for the response.
    """


class SlackElement(BaseModel):
    """An element of a Slack Block."""

    type: Optional[str] = Field(description="Type of block")
    """Common types are:
        - rich_text_section
        - user
        - text
    """
    elements: Optional[List["SlackElement"]] = Field(description="Element")
    user_id: Optional[str] = Field(description="The user id that created the element")
    text: Optional[str] = Field(description="The text of the element. Does not contain the user")

    def to_blocks(self) -> Optional[List[Block]]:
        """Return self as a list of Steamship Blocks."""
        ret = []

        if self.elements:
            for element in self.elements or []:
                for block in element.to_blocks() or []:
                    ret.append(block)
        elif self.text:
            block = Block(text=self.text)
            ret.append(block)
        return ret


class SlackBlock(BaseModel):
    """A Slack Block."""

    type: Optional[str] = Field(description="Type of block")
    """Common types are:
        - rich_text
    """
    block_id: Optional[str] = Field(description="ID of the block")
    elements: Optional[List[SlackElement]] = Field(description="Element")

    def to_blocks(self) -> Optional[List[Block]]:
        """Return self as a list of Steamship Blocks."""

        ret = []
        for element in self.elements or []:
            for block in element.to_blocks() or []:
                ret.append(block)
        return ret


class SlackEvent(BaseModel):
    """A Slack Event as reported to the Bot."""

    type: Optional[str] = Field(description="Type of event")
    """Common types are:
        - app_home_opened
        - app_mention,
        - reaction_added
        - message.channels - A message was posted to a channel
        - message.groups - A message was posted to a private channel
        - message.im - A message was posted to a DM
        - message.mpim - A message was posted to a Group DM
        - message.app_home - A message was posted to the "App Home" chat
        - hello - The client has connected to the server

    See more at: https://api.slack.com/events

    """
    user: Optional[str] = Field(
        description="The user ID belonging to the user that incited this action. Not included in all events as not all events are controlled by users. This is not the Slack handle."
    )
    channel: Optional[str] = Field(description="Channel of the event")
    tab: Optional[str] = Field(description="Tab in which the event occurred")
    event_ts: Optional[str] = Field(
        description="Timestamp of the event. A string, but is a floating point number within that."
    )
    ts: Optional[str] = Field(
        description="Timestamp of the message. A string, but is a floating point number within that."
    )
    thread_ts: Optional[str] = Field(
        description="Timestamp of the thread this message is a part of, if any.  Same format as `ts`."
    )
    item: Optional[str] = Field(
        description="Data specific to the underlying object type being described."
    )
    client_msg_id: Optional[str] = Field(description="")
    text: Optional[str] = Field(
        description="The text preview of the sequence of messages. Contains the user ID."
    )
    blocks: Optional[List[SlackBlock]] = Field(
        description="The text preview of the sequence of messages"
    )
    team: Optional[str] = Field(description="The team from which the message came")
    bot_id: Optional[str] = Field(
        description="The unique identifier for the bot that sent the event."
    )

    def is_message(self) -> bool:
        """Return whether this event is a message"""
        return self.type and (self.type.startswith("message") or self.type == "app_mention")

    def to_blocks(self) -> Optional[List[Block]]:
        """Return self as a list of Steamship Blocks."""

        ret = []
        for slack_block in self.blocks or []:
            for block in slack_block.to_blocks() or []:
                if self.channel:
                    block.set_chat_id(self.channel)
                if self.ts:
                    block.set_message_id(self.ts)
                if self.user:
                    block.set_user_id(self.user)
                if self.thread_ts:
                    block.set_thread_id(self.thread_ts)
                ret.append(block)
        return ret


class SlackRequest(BaseModel):
    """The outer object presented when Slack sends a webhook notification.

    https://api.slack.com/apis/connections/events-api#event-type-structure
    """

    token: Optional[str] = Field(
        description="The shared-private callback token that authenticates this callback to the application as having come from Slack. Match this against what you were given when the subscription was created. If it does not match, do not process the event and discard it."
    )
    team_id: Optional[str] = Field(
        description="The unique identifier for the workspace/team where this event occurred."
    )

    api_app_id: Optional[str] = Field(
        description="The unique identifier for the application this event is intended for."
    )
    event: Optional[SlackEvent] = Field(description="Slack event")
    type: Optional[str] = Field(description="The type of inbound request")  # event_callback
    event_id: Optional[str] = Field(description="The event ID")
    event_time: Optional[int] = Field(description="The event time")
    event_context: Optional[str] = Field(description="The event context key")


class SlackTransportConfig(Config):
    """Configuration object for the SlackTransport."""

    # TODO: This is a placeholder for a V2 to include a choice of what kind of scope you wish your bot to have.
    #       For example: only reply to direct mentions? Listen to ALL messages?

    slack_api_base: str = Field(
        SLACK_API_BASE, description="Slack API base URL. If blank defaults to production Slack."
    )
    threading_behavior: SlackThreadingBehavior = Field(
        SlackThreadingBehavior.FOLLOW_THREADS.value,
        description="Whether the bot will always respond in threads, or only if the invocation was threaded",
    )
    context_behavior: SlackContextBehavior = Field(
        SlackContextBehavior.ENTIRE_CHANNEL.value,
        description="Whether the bot will be provided conversation context from the channel as a whole, or per thread.",
    )


class SlackTransport(Transport):
    # noqa: RST204
    """Slack Transport Mixin for Steamship Agent to Slack interaction.

    Current support:
    - Outputs: outputs: text, image, video, audio output
    - Inputs: text
    - Message-triggering events: direct mention, bot room entry

    The included manifest file below will register a Slack bot which can be added to rooms and communicated
    with directly. As configured below, it will not receive messages that were not intended for it.

    Integration flow:

    Slack needs a different style of integration than Telegram, involving a two-way handshake.

    The user:

    1) Clicks a "Bot Manifest Link" which is generated by your Steamship Agent Instance
       This manifest link contains requested access as well as callback URL information.
    2) Clicks "Accept" to create a Slack Bot using that manifest.
    3) Clicks "Install" to install the Slack Bot to a workspace, accepting the permissions
    4) Copy the Oauth Token from the "Settings > Install App" page.
    5) Set the Oauth token for your agent using the set_slack_access_token POST method

    At this point, (1) Slack knows about the Agent, and (2) the Agent knows about Slack.

    Interacting with the Bot on Slack will trigger a request/response loop in the Agent.
    """

    bot_token: Optional[str] = None
    agent_service: AgentService
    config: SlackTransportConfig

    def __init__(
        self,
        client: Steamship,
        config: SlackTransportConfig,
        agent_service: AgentService,
    ):
        super().__init__(client=client)
        self.bot_token = None
        self.agent_service = agent_service
        self.config = config

    def instance_init(self):
        """Called when the owning AgentService initializes for the first time."""
        pass

    def _manifest(self) -> dict:
        """Return the Slack Manifest which describes this app."""
        # When running in development, the below values will be none.
        invocable_instance_handle = (
            self.agent_service.context.invocable_instance_handle or "Development Steamship Agent"
        )
        invocable_handle = (
            self.agent_service.context.invocable_handle or "Development Steamship Agent"
        )

        # Slack only supports names of 25 characters or less
        invocable_instance_handle = invocable_instance_handle[:34]

        """Return the Slack Manifest for this Transport."""
        return {
            "display_information": {
                "name": invocable_instance_handle,
                "description": f"An instance of {invocable_handle} powered by Steamship",
            },
            "features": {
                "bot_user": {"display_name": invocable_instance_handle, "always_online": True}
            },
            "oauth_config": {
                "scopes": {
                    "bot": [
                        "commands",
                        "app_mentions:read",
                        "chat:write",
                        "chat:write.public",
                        "im:history",
                    ]
                }
            },
            "settings": {
                "org_deploy_enabled": True,
                "socket_mode_enabled": False,
                "token_rotation_enabled": False,
                "interactivity": {
                    "is_enabled": True,
                    "request_url": f"{self.agent_service.context.invocable_url}slack_respond",
                },
                "event_subscriptions": {
                    "bot_events": [
                        "app_home_opened",
                        "app_mention",
                        "message.app_home",
                        "message.im",
                    ],
                    "request_url": f"{self.agent_service.context.invocable_url}slack_event",
                },
            },
        }

    def _manifest_url(self) -> str:
        """Return the Manifest Installation URL of the Bot."""
        manifest_dict = self._manifest()
        manifest_json = json.dumps(manifest_dict)
        manifest_str = urllib.parse.quote(manifest_json)
        return f"https://api.slack.com/apps?new_app=1&manifest_json={manifest_str}"

    @get("slack_install_link", public=True)
    def slack_install_link(self) -> str:
        """Return the link to install this Agent as a Slack bot."""
        return self._manifest_url()

    def _send(self, blocks: List[Block], metadata: Metadata):  # noqa: C901
        """Send a response to the Slack chat.

        Slack uses the `text` field for fallback.
        But we want to also provide the blocks and attachments for multiple messages that are multi-modal
        """
        text = None
        slack_blocks = []
        chat_id = None
        thread_ts = None
        for block in blocks:
            # This is required for the public_url creation below.

            block.client = self.client

            if block.chat_id:
                chat_id = block.chat_id

            if block.thread_id:
                thread_ts = block.thread_id

            if block.is_text() or block.text:
                if not text:
                    # This is the fallback for mobile notifications
                    text = block.text
                # The real thing we send is a block
                if blocks is None:
                    blocks = []
                slack_blocks.append(
                    {"type": "section", "text": {"type": "mrkdwn", "text": block.text}}
                )
            elif block.is_image():
                image_url = block.to_public_url()
                slack_blocks.append(
                    {"type": "image", "image_url": image_url, "alt_text": image_url}
                )
            elif block.is_audio():
                audio_url = block.to_public_url()
                slack_blocks.append(
                    {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"<{audio_url}|Audio Message>",
                        },
                    }
                )
            elif block.is_video():
                video_url = block.to_public_url()
                slack_blocks.append(
                    {"type": "video", "video_url": video_url, "alt_text": video_url}
                )
            else:
                logging.error(
                    f"Slack transport unable to send a block of MimeType {block.mime_type}"
                )

        bot_token = self.get_slack_access_token()
        if bot_token is None:
            logging.error("Unable to send Slack Message: Slack transport had null bot token")
            return

        if not chat_id:
            logging.error("Unable to send Slack Message: no chat_id on output blocks")
            return

        headers = {
            "Authorization": f"Bearer {bot_token}",
            "Content-Type": "application/json",
        }
        body = {
            "blocks": slack_blocks,
            "text": text,  # This is for mobile previews. The "block" key has the real content.
            "channel": chat_id,
        }
        if thread_ts:
            body["thread_ts"] = thread_ts

        post_url = f"{self.config.slack_api_base}chat.postMessage"

        requests.post(
            post_url,
            headers=headers,
            json=body,
        )

    def build_emit_func(
        self, chat_id: str, incoming_message_ts: str, thread_ts: Optional[str]
    ) -> EmitFunc:
        """Return an EmitFun that sends messages to the appropriate Slack channel."""
        if self.config.threading_behavior == SlackThreadingBehavior.FOLLOW_THREADS.value:
            reply_thread_ts = thread_ts
        elif self.config.threading_behavior == SlackThreadingBehavior.ALWAYS_THREADED.value:
            reply_thread_ts = thread_ts or incoming_message_ts
        else:
            raise ValueError(f"Unhandled threading behavior: {self.config.threading_behavior}")

        def new_emit_func(blocks: List[Block], metadata: Metadata):
            for block in blocks:
                block.set_chat_id(chat_id)
                if reply_thread_ts:
                    block.set_thread_id(reply_thread_ts)
            return self.send(blocks, metadata)

        return new_emit_func

    def _get_context_id_for_response(self, channel: str, thread_ts: Optional[str]) -> str:
        if self.config.context_behavior == SlackContextBehavior.ENTIRE_CHANNEL.value:
            return channel
        elif (
            self.config.context_behavior == SlackContextBehavior.THREADS_ARE_NEW_CONVERSATIONS.value
        ):
            return channel if not thread_ts else f"{channel}-{thread_ts}"
        else:
            raise ValueError(f"Unhandled context behavior: {self.config.context_behavior}")

    def _respond_to_block(self, incoming_message: Block):
        """Respond to a single inbound message from Slack, posting the response back to Slack."""
        try:
            chat_id = incoming_message.chat_id
            thread_ts = incoming_message.thread_id
            context_id = self._get_context_id_for_response(chat_id, thread_ts)
            context = self.agent_service.build_default_context(context_id=context_id)

            context.chat_history.append_user_message(
                text=incoming_message.text, tags=incoming_message.tags
            )

            context.metadata["slack"] = {
                "channel": chat_id,
                "message_ts": incoming_message.message_id,
            }
            if thread_ts:
                context.metadata["slack"]["thread_ts"] = thread_ts

            # TODO: For truly async support, this emit fn will need to be wired in at the Agent level.
            context.emit_funcs = [
                self.build_emit_func(
                    chat_id=chat_id,
                    incoming_message_ts=incoming_message.message_id,
                    thread_ts=thread_ts,
                )
            ]

            # Add an LLM to the context, using the Agent's if it exists.
            llm = None
            agent = self.agent_service.get_default_agent()
            if hasattr(agent, "llm"):
                llm = agent.llm
            else:
                llm = OpenAI(client=self.client)

            context = with_llm(context=context, llm=llm)

            self.agent_service.run_agent(agent, context)

        except BaseException as e:
            logging.error(e)
            if chat_id is not None:
                response = self.response_for_exception(e, chat_id=chat_id)
                if chat_id is not None:
                    self.send([response])

    def _parse_inbound(self, payload: dict, context: Optional[dict] = None) -> Optional[Block]:
        """Note: This is required by the superclass, but the return value of the interface (Block) doesn't match
        what Slack will send (List[Block]). A longer term TODO is refactor the base class, but until the Slack class
        is of acceptable quality, it feels like that would be putting too much into one PR."""
        return None

    @post("slack_respond_sync", public=True)
    def slack_respond_sync(self, **kwargs) -> InvocableResponse[str]:  # noqa: C901
        """Respond to inbound Slack events. This is a PUBLIC endpoint."""
        try:
            slack_request = SlackRequest.parse_obj(kwargs)
            # TODO: For truly async requests, we'll have to find some way to plumb through the token.
            # For now it appears to be provided upon each inbound request.
            if slack_request.event:
                if slack_request.event.bot_id is None:
                    if slack_request.event.is_message():
                        log_message = f"User {slack_request.event.user} sent message in channel {slack_request.event.channel}"
                        if slack_request.event.thread_ts:
                            log_message += f" from within thread {slack_request.event.thread_ts}"
                        logging.info(log_message)
                        incoming_messages = slack_request.event.to_blocks()
                        for incoming_message in incoming_messages:
                            if incoming_message is not None:
                                logging.info(f"Responding to {incoming_message}")
                                self._respond_to_block(incoming_message)
                    elif slack_request.event.type == "app_home_opened":
                        logging.info(
                            f"User {slack_request.event.user} opened App Tab with channel {slack_request.event.channel}"
                        )
                        result = Block(text="Hi there!")
                        result.set_chat_id(slack_request.event.channel)
                        self.send([result])

        except BaseException as e:
            logging.error(e)
            chat_id = None
            if chat_id is not None:
                response = self.response_for_exception(e, chat_id=chat_id)
                if chat_id is not None:
                    self.send([response])

        # Even if we do nothing, make sure we return ok
        return InvocableResponse(string="OK")

    @post("slack_event", public=True)
    def slack_event(self, **kwargs) -> InvocableResponse[str]:
        """Respond to an inbound event from Slack."""
        task = self.agent_service.invoke_later("slack_respond_sync", arguments=kwargs)
        logging.info(
            f"/slack_event: Created task {task.task_id} to respond to inbound Slack Message {kwargs}"
        )
        return InvocableResponse(string="OK")

    @post("slack_respond", public=True)
    def slack_respond(self, **kwargs) -> InvocableResponse[str]:
        """Respond to an inbound event from Slack."""
        task = self.agent_service.invoke_later("slack_respond_sync", arguments=kwargs)
        logging.info(
            f"/slack_respond: Created task {task.task_id} to respond to inbound Slack Message {kwargs}"
        )
        return InvocableResponse(string="OK")

    @post("set_slack_access_token")
    def set_slack_access_token(self, token: str) -> InvocableResponse[str]:
        """Set the slack access token."""
        kv = KeyValueStore(
            client=self.agent_service.client, store_identifier=self.setting_store_key()
        )
        kv.set("slack_token", {"token": token})
        return InvocableResponse(string="OK")

    def get_slack_access_token(self) -> Optional[str]:
        """Return the Slack Access token, which permits the agent to post to Slack."""
        if self.bot_token:
            return self.bot_token
        kv = KeyValueStore(
            client=self.agent_service.client, store_identifier=self.setting_store_key()
        )
        v = kv.get("slack_token")
        if not v:
            return None
        self.bot_token = v.get("token", None)
        return self.bot_token

    @post("is_slack_token_set")
    def is_slack_token_set(self) -> InvocableResponse[bool]:
        """Return whether the Slack token has been set as a way for a remote UI to check status."""
        token = self.get_slack_access_token()
        if token is None:
            return InvocableResponse(json=False)
        return InvocableResponse(json=True)

    def setting_store_key(self):
        return f"{SETTINGS_KVSTORE_KEY}-{self.agent_service.context.invocable_instance_handle}"



import pytest
import requests
from assets.packages.transports.mock_telegram_api import MockTelegramApi
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import File, PackageInstance, Steamship
from steamship.agents.mixins.transports.telegram import TelegramTransport, TelegramTransportConfig
from steamship.agents.service.agent_service import AgentService

config_template = {
    "telegram_token": {"type": "string", "default": ""},
    "telegram_api_base": {"type": "string", "default": ""},
    "slack_api_base": {"type": "string", "default": ""},
}


def test_telegram_api_base():
    with Steamship.temporary_workspace() as client:
        transport = TelegramTransport(
            client=client,
            agent_service=AgentService(client=client),
            config=TelegramTransportConfig(),
        )
        transport.bot_token = "TOKEN"  # noqa: S105
        assert transport.get_api_root() == "https://api.telegram.org/botTOKEN"


@pytest.mark.usefixtures("client")
def test_telegram(client: Steamship):

    mock_telegram_api_path = PACKAGES_PATH / "transports" / "mock_telegram_api.py"
    telegram_agent_path = PACKAGES_PATH / "transports" / "test_transports_agent.py"

    with deploy_package(client, mock_telegram_api_path, wait_for_init=True) as (
        _,
        _,
        mock_chat_api,
    ):
        # Test that the default instance doesn't have an API set.

        # Removing the / from the invocation URL and passing it as the telegram_token allows
        # the bot token to not be empty and the two appended to each other to just equal the
        # invocation url.
        instance_config = {
            "telegram_token": "abcdefg",
            "telegram_api_base": mock_chat_api.invocation_url[:-1],
            "slack_api_base": mock_chat_api.invocation_url,
        }

        # Should be able to invoke setWebhook publicly
        response = requests.get(
            url=mock_chat_api.invocation_url + "/setWebhook",
            params={"url": "test", "allowed_updates": "", "drop_pending_updates": True},
        )
        assert response.ok
        # Test that the call to setWebhook wrote a file
        files = File.query(client, f'kind "{MockTelegramApi.WEBHOOK_TAG}"').files
        assert len(files) == 1
        assert files[0].tags[0].name == "test"

        files[0].delete()

        files = File.query(client, f'kind "{MockTelegramApi.WEBHOOK_TAG}"').files
        assert len(files) == 0

        with deploy_package(
            client,
            telegram_agent_path,
            version_config_template=config_template,
            instance_config=instance_config,
            wait_for_init=True,
        ) as (agent_package, agent_version, agent_instance):
            respond_method = "telegram_respond"

            # The configuration provided a token, so the token should be reported as having been set.
            assert agent_instance.invoke("is_telegram_token_set") is True

            # Test that agent called instance_init and registered webhook
            files = File.query(client, f'kind "{MockTelegramApi.WEBHOOK_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].name == agent_instance.invocation_url + respond_method

            files[0].delete()

            files = File.query(client, f'kind "{MockTelegramApi.WEBHOOK_TAG}"').files
            assert len(files) == 0

            # test sending messages (without auth)
            response = requests.post(
                url=agent_instance.invocation_url + f"/{respond_method}",
                json=MockTelegramApi.generate_inbound_webhook_body("a test"),
            )
            assert response.ok
            files = File.query(client, f'kind "{MockTelegramApi.TEXT_MESSAGE_TAG}"').files
            assert len(files) == 1

            assert (
                files[0].tags[0].name == "Response to: a test"
            )  # bug somewhere - results being url encoded in some envs

            assert files[0].tags[0].value == {MockTelegramApi.CHAT_ID_KEY: 1}

            # test sending another message; this has been a problem before
            agent_instance.invoke(
                respond_method, **MockTelegramApi.generate_inbound_webhook_body("another test")
            )
            files = File.query(client, f'kind "{MockTelegramApi.TEXT_MESSAGE_TAG}"').files
            assert len(files) == 2
            allowed_responses = ["Response to: a test", "Response to: another test"]
            for file in files:
                assert file.tags[0].name in allowed_responses
                assert file.tags[0].value == {MockTelegramApi.CHAT_ID_KEY: 1}

            # Test the agent sending a "photo"
            agent_instance.invoke(
                respond_method, **MockTelegramApi.generate_inbound_webhook_body("image")
            )
            files = File.query(client, f'kind "{MockTelegramApi.PHOTO_MESSAGE_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].value == {
                MockTelegramApi.CHAT_ID_KEY: 1,
                MockTelegramApi.PHOTO_KEY: "some image bytes",
            }

            # Test the agent sending "audio"
            agent_instance.invoke(
                respond_method, **MockTelegramApi.generate_inbound_webhook_body("audio")
            )
            files = File.query(client, f'kind "{MockTelegramApi.AUDIO_MESSAGE_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].value == {
                MockTelegramApi.CHAT_ID_KEY: 1,
                MockTelegramApi.AUDIO_KEY: "some audio bytes",
            }

            # Test the agent sending a "video"
            agent_instance.invoke(
                respond_method, **MockTelegramApi.generate_inbound_webhook_body("video")
            )
            files = File.query(client, f'kind "{MockTelegramApi.VIDEO_MESSAGE_TAG}"').files
            assert len(files) == 1
            assert files[0].tags[0].value == {
                MockTelegramApi.CHAT_ID_KEY: 1,
                MockTelegramApi.VIDEO_KEY: "some video bytes",
            }

            # Verify that we don't have any webhook set files
            files = File.query(client, f'kind "{MockTelegramApi.WEBHOOK_TAG}"').files
            assert len(files) == 0

            agent_instance.invoke("set_telegram_access_token", token="foo-bar")  # noqa: S106

            # Test that this triggered a web hook reset
            files = File.query(client, f'kind "{MockTelegramApi.WEBHOOK_TAG}"').files
            assert len(files) == 1

            files[0].delete()

            # Create another instance to test LATE BOUND Telegram connections.

            instance_config_without_token = {
                "telegram_api_base": mock_chat_api.invocation_url[:-1],
                "slack_api_base": mock_chat_api.invocation_url,
            }

            package_instance_2 = PackageInstance.create(
                client,
                package_id=agent_package.id,
                package_version_id=agent_version.id,
                config=instance_config_without_token,
            )

            package_instance_2.wait_for_init()

            # The configuration provided a token, so the token should be reported as having been set.
            assert package_instance_2.invoke("is_telegram_token_set") is False

            package_instance_2.invoke(
                "set_telegram_access_token", token=instance_config.get("telegram_token")
            )

            assert package_instance_2.invoke("is_telegram_token_set") is True

            # See that the webhook was registered
            files = File.query(client, f'kind "{MockTelegramApi.WEBHOOK_TAG}"').files
            assert len(files) == 1



from __future__ import annotations

import json
from typing import Any, Dict, Optional, Type

from pydantic import BaseModel, Field

from steamship.base.client import Client
from steamship.base.model import CamelModel
from steamship.base.request import IdentifierRequest, Request


class CreatePackageVersionRequest(Request):
    package_id: str = None
    handle: str = None
    type: str = "file"
    hosting_handler: str = None
    # Note: this is a Dict[str, Any] but should be transmitted to the Engine as a JSON string
    config_template: str = None


class PackageVersion(CamelModel):
    client: Client = Field(None, exclude=True)
    id: str = None
    package_id: str = None
    handle: str = None
    config_template: Dict[str, Any] = None

    @classmethod
    def parse_obj(cls: Type[BaseModel], obj: Any) -> BaseModel:
        # TODO (enias): This needs to be solved at the engine side
        obj = obj["packageVersion"] if "packageVersion" in obj else obj
        return super().parse_obj(obj)

    @staticmethod
    def create(
        client: Client,
        package_id: Optional[str] = None,
        handle: Optional[str] = None,
        filename: Optional[str] = None,
        filebytes: Optional[bytes] = None,
        config_template: Optional[Dict[str, Any]] = None,
        hosting_handler: Optional[str] = None,
    ) -> PackageVersion:

        if filename is None and filebytes is None:
            raise Exception("Either filename or filebytes must be provided.")
        if filename is not None and filebytes is not None:
            raise Exception("Only either filename or filebytes should be provided.")

        if filename is not None:
            with open(filename, "rb") as f:
                filebytes = f.read()

        req = CreatePackageVersionRequest(
            handle=handle,
            package_id=package_id,
            config_template=json.dumps(config_template or {}),
            hosting_handler=hosting_handler,
        )

        task = client.post(
            "package/version/create",
            payload=req,
            file=("package.zip", filebytes, "multipart/form-data"),
            expect=PackageVersion,
        )
        task.wait_until_completed()
        return task.output

    def delete(self) -> PackageVersion:
        """Delete this package version. If this package is public and another user has created an instance of this version,
        this method will throw an error and the PackageVersion will not be deleted. Deleting the PackageVersion will delete any
        instances of it that you have created."""
        return self.client.post(
            "package/version/delete",
            IdentifierRequest(id=self.id),
            expect=PackageVersion,
        )



import logging
import tempfile
from typing import Any, Dict, List, Optional

import requests
from pydantic import Field

from steamship import Block, Steamship, SteamshipError
from steamship.agents.mixins.transports.transport import Transport
from steamship.agents.schema import EmitFunc, Metadata
from steamship.agents.service.agent_service import AgentService
from steamship.invocable import Config, InvocableResponse, post
from steamship.utils.kv_store import KeyValueStore

SETTINGS_KVSTORE_KEY = "telegram-transport"


class TelegramTransportConfig(Config):
    bot_token: Optional[str] = Field("", description="The secret token for your Telegram bot")
    api_base: Optional[str] = Field(
        "https://api.telegram.org/bot", description="The root API for Telegram"
    )


class TelegramTransport(Transport):
    """Experimental base class to encapsulate a Telegram communication channel."""

    bot_token: Optional[str] = None
    agent_service: AgentService
    config: TelegramTransportConfig

    def __init__(
        self,
        client: Steamship,
        config: TelegramTransportConfig,
        agent_service: AgentService,
    ):
        super().__init__(client=client)
        self.config = config
        self.agent_service = agent_service
        try:
            self.bot_token = self.get_telegram_access_token() or None
        except BaseException as e:
            logging.warning(e)
            self.bot_token = None

    def instance_init(self):
        if self.get_telegram_access_token():
            try:
                self.telegram_connect_webhook()
            except Exception:  # noqa: S110
                pass

    @post("telegram_connect_webhook")
    def telegram_connect_webhook(self):
        """Register this AgentService with Telegram."""
        webhook_url = self.agent_service.context.invocable_url + "telegram_respond"

        api_root = self.get_api_root()
        if not api_root:
            raise SteamshipError(
                message="Unable to determine Telegram API root -- perhaps your bot token isn't set?"
            )

        logging.info(
            f"Setting Telegram webhook URL: {webhook_url}. Post is to {api_root}/setWebhook"
        )

        response = requests.get(
            f"{api_root}/setWebhook",
            params={
                "url": webhook_url,
                "allowed_updates": ["message"],
                "drop_pending_updates": True,
            },
        )

        if not response.ok:
            raise SteamshipError(
                f"Could not set webhook for bot. Webhook URL was {webhook_url}. Telegram response message: {response.text}"
            )

    @post("telegram_webhook_info")
    def telegram_webhook_info(self) -> dict:
        api_root = self.get_api_root()
        if not api_root:
            raise SteamshipError(
                message="Unable to fetch Telegram API info -- perhaps your bot token isn't set?"
            )

        return requests.get(api_root + "/getWebhookInfo").json()

    @post("telegram_disconnect_webhook")
    def telegram_disconnect_webhook(self, *args, **kwargs):
        """Unsubscribe from Telegram updates."""
        api_root = self.get_api_root()
        if not api_root:
            raise SteamshipError(
                message="Unable to disconnect from Telegram -- perhaps your bot token isn't set?"
            )

        requests.get(f"{api_root}/deleteWebhook")

    def _send(self, blocks: [Block], metadata: Metadata):
        """Send a response to the Telegram chat."""
        api_root = self.get_api_root()
        if not api_root:
            raise SteamshipError(
                message="Unable to send to Telegram -- perhaps your bot token isn't set?"
            )

        for block in blocks:
            chat_id = block.chat_id
            if block.is_text() or block.text:
                params = {"chat_id": int(chat_id), "text": block.text}
                requests.get(f"{api_root}/sendMessage", params=params)
            elif block.is_image() or block.is_audio() or block.is_video():
                if block.is_image():
                    suffix = "sendPhoto"
                    key = "photo"
                elif block.is_audio():
                    suffix = "sendAudio"
                    key = "audio"
                elif block.is_video():
                    suffix = "sendVideo"
                    key = "video"

                _bytes = block.raw()
                with tempfile.TemporaryFile(mode="r+b") as temp_file:
                    temp_file.write(_bytes)
                    temp_file.seek(0)
                    resp = requests.post(
                        url=f"{api_root}/{suffix}?chat_id={chat_id}",
                        files={key: temp_file},
                    )
                    if resp.status_code != 200:
                        logging.error(f"Error sending message: {resp.text} [{resp.status_code}]")
                        raise SteamshipError(
                            f"Message not sent to chat {chat_id} successfully: {resp.text}"
                        )
            else:
                logging.error(
                    f"Telegram transport unable to send a block of MimeType {block.mime_type}"
                )

    def _get_file(self, file_id: str) -> Dict[str, Any]:
        api_root = self.get_api_root()
        if not api_root:
            raise SteamshipError(
                message="Unable to get Telegram file -- perhaps your bot token isn't set?"
            )

        return requests.get(f"{api_root}/getFile", params={"file_id": file_id}).json()["result"]

    def _get_file_url(self, file_id: str) -> str:
        return f"https://api.telegram.org/file/bot{self.get_telegram_access_token()}/{self._get_file(file_id)['file_path']}"

    def _download_file(self, file_id: str):
        result = requests.get(self._get_file_url(file_id))
        if result.status_code != 200:
            raise Exception("Download file", result)

        return result.content

    def _parse_inbound(self, payload: dict, context: Optional[dict] = None) -> Optional[Block]:
        """Parses an inbound Telegram message."""

        chat = payload.get("chat")
        if chat is None:
            raise SteamshipError(f"No `chat` found in Telegram message: {payload}")

        chat_id = chat.get("id")
        if chat_id is None:
            raise SteamshipError(f"No 'chat_id' found in Telegram message: {payload}")

        if not isinstance(chat_id, int):
            raise SteamshipError(
                f"Bad 'chat_id' found in Telegram message: ({chat_id}). Should have been an int."
            )

        message_id = payload.get("message_id")
        if message_id is None:
            raise SteamshipError(f"No 'message_id' found in Telegram message: {payload}")

        if not isinstance(message_id, int):
            raise SteamshipError(
                f"Bad 'message_id' found in Telegram message: ({message_id}). Should have been an int"
            )

        if video_or_voice := (payload.get("voice") or payload.get("video_note")):
            file_id = video_or_voice.get("file_id")
            file_url = self._get_file_url(file_id)
            block = Block(
                text=payload.get("text"),
                url=file_url,
            )
            block.set_chat_id(str(chat_id))
            block.set_message_id(str(message_id))
            return block

        # Some incoming messages (like the group join message) don't have message text.
        # Rather than throw an error, we just don't return a Block.
        message_text = payload.get("text")
        if message_text is not None:
            result = Block(text=message_text)
            result.set_chat_id(str(chat_id))
            result.set_message_id(str(message_id))
            return result
        else:
            return None

    def build_emit_func(self, chat_id: str) -> EmitFunc:
        def new_emit_func(blocks: List[Block], metadata: Metadata):
            for block in blocks:
                block.set_chat_id(chat_id)
            return self.send(blocks, metadata)

        return new_emit_func

    @post("telegram_respond", public=True)
    def telegram_respond(self, **kwargs) -> InvocableResponse[str]:
        """Endpoint implementing the Telegram WebHook contract. This is a PUBLIC endpoint since Telegram cannot pass a Bearer token."""

        # TODO: must reject things not from the package
        message = kwargs.get("message", {})
        chat_id = message.get("chat", {}).get("id")
        try:
            incoming_message = self.parse_inbound(message)
            if incoming_message is not None:
                context = self.agent_service.build_default_context(chat_id)

                context.chat_history.append_user_message(
                    text=incoming_message.text, tags=incoming_message.tags
                )
                context.emit_funcs = [self.build_emit_func(chat_id=chat_id)]

                response = self.agent_service.run_agent(
                    self.agent_service.get_default_agent(), context
                )
                if response is not None:
                    self.send(response)
                else:
                    # Do nothing here; this could be a message we intentionally don't want to respond to (ex. an image or file upload)
                    pass
            else:
                # Do nothing here; this could be a message we intentionally don't want to respond to (ex. an image or file upload)
                pass
        except Exception as e:
            response = self.response_for_exception(e, chat_id=chat_id)

            if chat_id is not None:
                self.send([response])
        # Even if we do nothing, make sure we return ok
        return InvocableResponse(string="OK")

    @post("set_telegram_access_token")
    def set_telegram_access_token(self, token: str) -> InvocableResponse[str]:
        """Set the telegram access token."""
        existing_token = self.get_telegram_access_token()
        if existing_token:
            try:
                self.telegram_disconnect_webhook()
            except BaseException as e:
                # Note: we don't want to fully fail here, because that would mean that a bot user that had some
                # other error relating to disconnecting would never be able to RE-connect to a new bot.
                logging.error(e)

        kv = KeyValueStore(
            client=self.agent_service.client, store_identifier=self.setting_store_key()
        )
        kv.set("telegram_token", {"token": token})

        # Now attempt to modify the connection in Telegram
        self.bot_token = token
        try:
            self.telegram_connect_webhook()
            return InvocableResponse(string="OK")
        except Exception as e:
            raise SteamshipError(message=f"Could not set Telegram Webhook. Exception: {e}")

    def get_api_root(self) -> Optional[str]:
        """Return the API root"""
        bot_token = self.get_telegram_access_token()
        api_base = self.config.api_base

        # Ensure we have an API Base
        if not api_base:
            raise SteamshipError(message="Missing Telegram API Base")

        # Ensure it ends in a trailing slash
        if api_base[-1] != "/":
            api_base += "/"

        if bot_token:
            if ".steamship.run/" in api_base or ".apps.staging.steamship.com" in api_base:
                # This is a special case for our testing pipeline -- it contains a mock Telegram server.
                return api_base
            else:
                return f"{api_base[:-1]}{bot_token}"
        else:
            return None

    def setting_store_key(self):
        return f"{SETTINGS_KVSTORE_KEY}-{self.agent_service.context.invocable_instance_handle}"

    def get_telegram_access_token(self) -> Optional[str]:
        """Return the Telegram Access token, which permits the agent to post to Telegram."""

        # Warning: This can't be an 'is not None' check since the config system uses an empty string to represent None
        if self.bot_token:
            return self.bot_token

        _dynamically_set_token = None
        _fallback_token = None

        # Prefer the dynamically set token if available
        kv = KeyValueStore(
            client=self.agent_service.client, store_identifier=self.setting_store_key()
        )
        v = kv.get("telegram_token")
        if v:
            _dynamically_set_token = v.get("token", None)

        # Fall back on the config-provided one
        if self.config:
            _fallback_token = self.config.bot_token

        _return_token = _dynamically_set_token or _fallback_token

        # Cache it to avoid another KV Store lookup and return
        self.bot_token = _return_token
        return _return_token

    @post("is_telegram_token_set")
    def is_telegram_token_set(self) -> InvocableResponse[bool]:
        """Return whether the Telegram token has been set as a way for a remote UI to check status."""

        # Warning: This can't be an 'is not None' check since the config system uses an empty string to represent None
        if not self.get_telegram_access_token():
            return InvocableResponse(json=False)
        return InvocableResponse(json=True)



from abc import ABC, abstractmethod
from typing import List, Optional

from pydantic.main import BaseModel

from steamship import Block
from steamship.agents.schema.tool import Tool


class LLM(BaseModel, ABC):
    """LLM wraps large language model-based backends.

    They may be used with LLMAgents in Action selection, or for direct prompt completion."""

    @abstractmethod
    def complete(self, prompt: str, stop: Optional[str] = None, **kwargs) -> List[Block]:
        """Completes the provided prompt, stopping when the stop sequeunce is found."""
        pass


# TODO(dougreid): should LLM and ConversationalLLM share a common parent?
class ChatLLM(BaseModel, ABC):
    """ChatLLM wraps large language model-based backends that use a chat completion style interation.

    They may be used with Agents in Action selection, or for direct prompt completion."""

    @abstractmethod
    def chat(self, messages: List[Block], tools: Optional[List[Tool]], **kwargs) -> List[Block]:
        """Sends the set of chat messages to the LLM, returning the next part of the conversation"""
        pass



import hashlib
import json
import logging
from typing import Dict, List, Optional

from steamship import Block, MimeTypes, Steamship
from steamship.agents.schema.action import Action, FinishAction
from steamship.utils.kv_store import KeyValueStore


def _blocks_to_cache_dict(value: List[Block]) -> Dict[str, any]:
    """Attempts to convert a list of blocks to key-store-safe dictionary.

    Convention: {'blocks':[{'id': <block_id>}, {'text': <some_text>}]}
    """

    if not value:
        return {}
    if not isinstance(value, list):
        return {}
    blocks = []
    for block in value:
        if block.id:
            blocks.append({"id": block.id})
        else:
            # TODO(dougreid): safe assumption about temporary blocks?
            blocks.append({"text": block.text})
    return {"blocks": blocks}


def _blocks_from_cache_dict(client: Steamship, value: Dict[str, any]) -> List[Block]:
    """Attempts to convert a key-store-safe dictionary to a list of blocks.

    Convention: {'blocks':[{'id': <block_id>}, {'text': <some_text>}]}
    """
    if block_list := value.get("blocks"):
        return_blocks = []
        for b in block_list:
            if block_id := b.get("id"):
                return_blocks.append(Block.get(client, _id=block_id))
            else:
                return_blocks.append(Block(text=b.get("text"), mime_type=MimeTypes.TXT))
        return return_blocks

    return []


class ActionCache:
    """Provide persistent cache layer for AgentContext that allows lookups of output blocks from Actions.

    Use this cache to eliminate calls to Tools.

    NOTE: EXPERIMENTAL.
    """

    client: Steamship
    key_value_store: KeyValueStore

    def __init__(self, client: Steamship, key_value_store: KeyValueStore):
        self.client = client
        self.key_value_store = key_value_store

    @staticmethod
    def get_or_create(client: Steamship, context_keys: Dict[str, str]):
        cache_handle = (
            f"actioncache-{hashlib.sha256(json.dumps(context_keys).encode('utf-8')).hexdigest()}"
        )
        return ActionCache(
            client=client,
            key_value_store=KeyValueStore(client=client, store_identifier=cache_handle),
        )

    @staticmethod
    def _cache_key_for(action: Action) -> str:
        if isinstance(action, Action):
            key = f"{action.tool}"
            for block in action.input:
                if block.is_text():
                    key = f"{key}-{block.text}"
                else:
                    key = f"{key}-{block.id}"
        else:
            key = f"unknown-{json.dumps(action)}"

        return f"sk-{hashlib.sha256(key.encode('utf-8')).hexdigest()}"

    def lookup(self, key: Action) -> Optional[List[Block]]:
        cache_key = ActionCache._cache_key_for(key)
        value = self.key_value_store.get(key=cache_key) or None
        if value:
            logging.debug(f"cache hit for {cache_key}")
            return _blocks_from_cache_dict(self.client, value)

        logging.debug(f"cache miss for {cache_key}")
        return None

    def clear(self) -> None:
        self.key_value_store.reset()

    def update(self, key: Action, value: List[Block]):
        # TODO: should this be synchronous and wait?
        self.key_value_store.set(
            key=ActionCache._cache_key_for(key), value=_blocks_to_cache_dict(value)
        )
        return

    def delete(self, key: Action) -> bool:
        return self.key_value_store.delete(key=ActionCache._cache_key_for(key))


class LLMCache:
    """Provide persistent cache layer for AgentContext that allows lookups of Actions from LLM prompts.

    Use this cache to eliminate calls to LLMs for Tool selection and direct responses.

    NOTE: EXPERIMENTAL.
    """

    client: Steamship
    key_value_store: KeyValueStore

    def __init__(self, client: Steamship, key_value_store: KeyValueStore):
        self.client = client
        self.key_value_store = key_value_store

    @staticmethod
    def get_or_create(client: Steamship, context_keys: Dict[str, str]):
        cache_handle = (
            f"llmcache-{hashlib.sha256(json.dumps(context_keys).encode('utf-8')).hexdigest()}"
        )
        return LLMCache(
            client=client,
            key_value_store=KeyValueStore(client=client, store_identifier=cache_handle),
        )

    @staticmethod
    def _cache_key_for(inputs: List[Block]) -> str:
        if isinstance(inputs, list):
            key = "llm"
            for block in inputs:
                if block.is_text():
                    key = f"{key}-{block.text}"
                else:
                    key = f"{key}-{block.id}"
        else:
            key = f"unknown-{json.dumps(inputs)}"
        return f"sk-{hashlib.sha256(key.encode('utf-8')).hexdigest()}"

    def _action_from_value(self, value: dict) -> Action:
        input_blocks = []
        if in_blocks := value.get("input"):
            input_blocks = _blocks_from_cache_dict(self.client, in_blocks)

        output_blocks = []
        if out_blocks := value.get("output"):
            output_blocks = _blocks_from_cache_dict(self.client, out_blocks)

        if value.get("tool") == "Agent-FinishAction":
            return FinishAction(input=input_blocks, output=output_blocks)

        return Action(tool=value.get("tool"), input=input_blocks, output=output_blocks)

    def lookup(self, key: List[Block]) -> Optional[Action]:
        cache_key = LLMCache._cache_key_for(key)
        value = self.key_value_store.get(key=cache_key) or None
        if value:
            logging.debug(f"cache hit for {cache_key}")
            return self._action_from_value(value)

        logging.debug(f"cache miss for {cache_key}")
        return None

    def clear(self) -> None:
        self.key_value_store.reset()

    def update(self, key: List[Block], value: Action):
        # TODO: should this be synchronous and wait?
        action_dict = {
            "tool": value.tool,
            "input": _blocks_to_cache_dict(value.input),
            "output": _blocks_to_cache_dict(value.output),
        }
        self.key_value_store.set(key=LLMCache._cache_key_for(key), value=action_dict)
        return

    def delete(self, key: List[Block]) -> bool:
        return self.key_value_store.delete(key=LLMCache._cache_key_for(key))



import time

from steamship import MimeTypes
from steamship.invocable import InvocableResponse
from steamship.plugin.inputs.raw_block_and_tag_plugin_input import RawBlockAndTagPluginInput
from steamship.plugin.inputs.raw_block_and_tag_plugin_input_with_preallocated_blocks import (
    RawBlockAndTagPluginInputWithPreallocatedBlocks,
)
from steamship.plugin.outputs.block_type_plugin_output import BlockTypePluginOutput
from steamship.plugin.outputs.stream_complete_plugin_output import StreamCompletePluginOutput
from steamship.plugin.request import PluginRequest
from steamship.plugin.streaming_generator import StreamingGenerator


class TestStreamingGenerator(StreamingGenerator):
    def run(
        self, request: PluginRequest[RawBlockAndTagPluginInputWithPreallocatedBlocks]
    ) -> InvocableResponse[StreamCompletePluginOutput]:

        chunk_delay = int((request.data.options or {}).get("chunk_delay", "100"))
        for i, block in enumerate(request.data.blocks):
            output_block = request.data.output_blocks[i]
            output_text = block.text[::-1]
            for chunk in TestStreamingGenerator.split_string_into_chunks(output_text, 2):
                output_block.append_stream(bytes=bytes(chunk, "utf-8"))
                time.sleep(chunk_delay / 1000)
            output_block.finish_stream()

        return InvocableResponse(data=StreamCompletePluginOutput())

    def determine_output_block_types(
        self, request: PluginRequest[RawBlockAndTagPluginInput]
    ) -> InvocableResponse[BlockTypePluginOutput]:
        # Output block types are one text block per input block
        result = [MimeTypes.TXT.value] * len(request.data.blocks)
        return InvocableResponse(data=BlockTypePluginOutput(block_types_to_create=result))

    @staticmethod
    def split_string_into_chunks(s: str, chunk_size: int):
        length = len(s)
        result = []
        for pos in range(0, length, chunk_size):
            result.append(s[pos : pos + chunk_size])
        return result



from steamship.invocable import InvocableResponse, PackageService, get, post

# This is reflected in the docs as Package Project Structure.  If you have to change it
# for SDK changes, make sure to update the docs as well.


class MyPackage(PackageService):
    @get("say_hello")
    def _method_name_need_not_match(self, name: str = None) -> InvocableResponse:
        return InvocableResponse(string=f"Hello, {name}")

    @post("do_something")
    def do_something(self, number: int = None) -> InvocableResponse:
        return InvocableResponse(json={"number": number})



import json

import pytest
import requests
from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import Steamship


@pytest.mark.usefixtures("client")
def test_instance_init_is_called_on_package(client: Steamship):
    demo_package_path = PACKAGES_PATH / "echo_test_header.py"

    with deploy_package(client, demo_package_path) as (_, _, instance):
        url = instance.invocation_url + "/echo_test_header"
        response = requests.get(
            url,
            headers={
                "authorization": f"bearer {client.config.api_key.get_secret_value()}",
                "test1": "testValue1",
                "test2": "testValue2",
            },
        )
        result = json.loads(response.text)
        assert result is not None
        assert result["test1"] == "testValue1"
        assert result["test2"] == "testValue2"



import json

from steamship.invocable import PackageService, get


class EchoTestHeader(PackageService):
    @get("echo_test_header")
    def echo_test_header(self) -> str:
        return json.dumps(self.context.headers)



import base64
import json

from steamship import Task
from steamship.invocable import InvocableResponse, PackageService, post


class RequestIdPlumbingTestPackage(PackageService):

    # A test package that simply returns the request ID that it was given
    # by the proxy as well as the request ID that is passed to the plugin.
    # These should be the same.

    @post("requestids")
    def request_ids(self, plugin_handle: str) -> InvocableResponse:
        generator = self.client.use_plugin(plugin_handle)
        generate_task = generator.generate(text="")
        self_task = self.invoke_later("justmyrequestid")
        plugin_request_id = generate_task.wait().blocks[0].text

        self_task_fetched = Task.get(
            self.client, _id=self_task.task_id
        )  # to undo the wrong generic Task output
        invoke_later_request_id_dict = json.loads(
            str(base64.b64decode(self_task_fetched.wait()), encoding="UTF-8")
        )  # invoke_later tasks get stored b64 encoded
        return InvocableResponse(
            json={
                "packageRequestId": self.client.config.request_id,
                "pluginRequestId": plugin_request_id,
                "invokeLaterRequestId": invoke_later_request_id_dict.get("requestId"),
            }
        )

    @post("justmyrequestid")
    def just_my_request_id(self) -> InvocableResponse:
        return InvocableResponse(json={"requestId": self.client.config.request_id})



from typing import Any, Callable, List, Optional, Union

import pytest

from steamship import Block, Task
from steamship.agents.functional import FunctionsBasedAgent
from steamship.agents.llms.openai import ChatOpenAI
from steamship.agents.schema import AgentContext, Tool
from steamship.agents.schema.message_selectors import MessageWindowMessageSelector
from steamship.agents.service.agent_service import AgentService


class NonFinalTool(Tool):
    name = "SuggestFood"
    agent_description = "Tool that suggests a food. Use whenever someone wants a food suggestion. Input: the food suggestion request. Output: the food suggested."
    human_description = "SuggestFood"
    is_final: bool = False

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        return [Block(text="apples")]


class FinalTool(Tool):
    name = "SuggestPlace"
    agent_description = "Tool that suggests a place. Use whenever someone wants a place suggestion. Input: the place suggestion request. Output: the place suggested."
    human_description = "SuggestPlace"
    is_final: bool = True

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        return [Block(text="Peru")]


class MyFinalToolAssistant(AgentService):
    """MyFinalToolAssistant is an example AgentService that exposes a single test endpoint
    for testing the `is_final` marker on Tools. It is configured with two simple Tool: one
    is final and one is not. It is instructed to have a strong personality which should it cause it
    to rewrite the tool output if the tool is not final, but not to rewrite the tool output if it
    is final."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs, use_llm_cache=True, use_action_cache=True)
        self.set_default_agent(
            FunctionsBasedAgent(
                tools=[
                    NonFinalTool(),
                    FinalTool(),
                ],
                llm=ChatOpenAI(self.client, temperature=0),
                conversation_memory=MessageWindowMessageSelector(k=2),
            )
        )


@pytest.mark.parametrize("invocable_handler", [MyFinalToolAssistant], indirect=True)
def test_final_tool_assistant(invocable_handler: Callable[[str, str, Optional[dict]], dict]):
    food_suggestion = invocable_handler("POST", "/prompt", {"prompt": "Suggest a food"}).get("data")
    assert food_suggestion
    assert food_suggestion[0].get("text")

    # The agent responded with a FULL sentence. Something like: "I suggest you try some apples!"
    assert food_suggestion[0].get("text") != "apples"
    assert "apples" in food_suggestion[0].get("text")

    # Now we'll ask for a place, which uses the is_final=True tool.
    # The response should be exactly and only the tool output.

    food_suggestion = invocable_handler("POST", "/prompt", {"prompt": "Suggest a place"}).get(
        "data"
    )
    assert food_suggestion
    assert food_suggestion[0].get("text")

    # The agent responded with a FULL sentence. Something like: "I suggest you try some apples!"
    assert food_suggestion[0].get("text") == "Peru"



from typing import List, Optional

from steamship.data.tags.tag import Tag


def get_tag(
    tags: Optional[List[Tag]], kind: Optional[str] = None, name: Optional[str] = None
) -> Optional[Tag]:
    """Return the first tag of a list with the provided kind & name."""
    for tag in tags or []:
        if (kind is None or tag.kind == kind) and (name is None or tag.name == name):
            return tag
    return None


def get_tag_value_key(
    tags: Optional[List[Tag]], key: str, kind: Optional[str] = None, name: Optional[str] = None
) -> Optional[any]:
    """Return the value key from the first tag of a list with the provided kind & name."""
    if tag := get_tag(tags, kind=kind, name=name):
        return (tag.value or {}).get(key)
    return None



import pytest

from steamship import Block, Steamship
from steamship.agents.llms import OpenAI
from steamship.agents.llms.openai import ChatOpenAI


@pytest.mark.usefixtures("client")
def test_openai_llm(client: Steamship):
    llm = OpenAI(client=client, temperature=0)
    blocks = llm.complete(prompt="Why did the chicken cross the road?", stop="side.")

    assert len(blocks) == 1
    assert blocks[0].is_text()


@pytest.mark.usefixtures("client")
def test_openai_llm_model_selection(client: Steamship):
    llm = OpenAI(client=client, temperature=0, model_name="gpt-4")
    blocks = llm.complete(prompt="Why did the chicken cross the road?", stop="side.")

    assert len(blocks) == 1
    assert blocks[0].is_text()


@pytest.mark.usefixtures("client")
def test_openai_llm_model_max_tokens_static(client: Steamship):
    llm = OpenAI(client=client, temperature=0, max_tokens=1)
    blocks = llm.complete(prompt="Why did the chicken cross the road?")

    assert len(blocks) == 1
    assert blocks[0].is_text()
    assert len(blocks[0].text.split()) == 1


@pytest.mark.usefixtures("client")
def test_openai_llm_model_max_tokens_dynamic(client: Steamship):
    llm = OpenAI(client=client, temperature=0)
    blocks = llm.complete(prompt="Why did the chicken cross the road?", max_tokens=1)

    assert len(blocks) == 1
    assert blocks[0].is_text()
    assert len(blocks[0].text.split()) == 1


@pytest.mark.usefixtures("client")
def test_openai_llm_chat_model_max_tokens_dynamic(client: Steamship):
    llm = ChatOpenAI(client=client, temperature=0)
    blocks = llm.chat(
        messages=[Block(text="Why did the chicken cross the road?")], tools=[], max_tokens=1
    )

    assert len(blocks) == 1
    assert blocks[0].is_text()
    assert len(blocks[0].text.split()) == 1
